from pbi_core.ssas.model_tables.base import SsasReadonlyRecord


class CalcDependency(SsasReadonlyRecord):
    """Calculation Dependency.

    Represents a dependency between two DAX calculations in the model.
    This is recursive, so it connects non-direct dependencies.
    For instance, if we have three measures (A -> B -> C) there will be a dependency record between A and C.
    This entity is calculated, rather than being "real" like the other entities.


    SSAS spec:
    """

    database_name: str
    object_type: str
    table: str | None = None
    object: str
    expression: str | None = None
    referenced_object_type: str
    referenced_table: str | None = None
    referenced_object: str
    referenced_expression: str | None = None

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        return f"{self.object_type}[{self.object}] -> {self.referenced_object_type}[{self.referenced_object}]"



import datetime
from typing import TYPE_CHECKING, Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.linguistic_metadata import LinguisticMetadata
    from pbi_core.ssas.model_tables.model import Model


class Culture(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/d3770118-47bf-4304-9edf-8025f4820c45)
    """

    linguistic_metadata_id: int
    model_id: int
    name: str

    modified_time: datetime.datetime
    structure_modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.culture)

    def linguistic_metdata(self) -> "LinguisticMetadata":
        return self.tabular_model.linguistic_metadata.find({"id": self.linguistic_metadata_id})

    def model(self) -> "Model":
        return self.tabular_model.model

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type, [self.linguistic_metdata().get_lineage(lineage_type)])
        return LineageNode(self, lineage_type, [self.model().get_lineage(lineage_type)])



import datetime
from typing import TYPE_CHECKING, Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.model_tables.enums import DataState
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import CrossFilteringBehavior, JoinOnDateBehavior, RelationshipType, SecurityFilteringBehavior

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.model import Model
    from pbi_core.ssas.model_tables.table import Table
    from pbi_core.ssas.model_tables.variation import Variation


class Relationship(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/35bb4a68-b97e-409b-a5dd-14695fd99139)
    This class represents a relationship between two tables in a Tabular model.
    """

    cross_filtering_behavior: CrossFilteringBehavior
    from_column_id: int
    from_cardinality: int
    from_table_id: int
    is_active: bool
    join_on_date_behavior: JoinOnDateBehavior
    model_id: int
    name: str
    relationship_storage_id: int | None = None
    relationship_storage2_id: int | None = None
    relationship_storage2id: int | None = None  # TODO: check which one is wrong
    rely_on_referential_integrity: bool
    security_filtering_behavior: SecurityFilteringBehavior
    state: DataState
    to_cardinality: int
    to_column_id: int
    to_table_id: int
    type: RelationshipType

    modified_time: datetime.datetime
    refreshed_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.relationship)

    def from_table(self) -> "Table":
        """Returns the table the relationship is using as a filter.

        Note:
            In the bi-directional case, this table is also filtered

        """
        return self.tabular_model.tables.find({"id": self.from_table_id})

    def to_table(self) -> "Table":
        """Returns the table the relationship is being filtered.

        Note:
            In the bi-directional case, this table is also used as a filter

        """
        return self.tabular_model.tables.find({"id": self.to_table_id})

    def from_column(self) -> "Column":
        """The column in the from_table used to join with the to_table."""
        return self.tabular_model.columns.find({"id": self.from_column_id})

    def to_column(self) -> "Column":
        """The column in the to_table used to join with the from_table."""
        return self.tabular_model.columns.find({"id": self.to_column_id})

    def model(self) -> "Model":
        """The DB model this entity exists in."""
        return self.tabular_model.model

    def variations(self) -> set["Variation"]:
        return self.tabular_model.variations.find_all({"relationship_id": self.id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [variation.get_lineage(lineage_type) for variation in self.variations()],
            )
        return LineageNode(
            self,
            lineage_type,
            [
                self.from_table().get_lineage(lineage_type),
                self.to_table().get_lineage(lineage_type),
                self.from_column().get_lineage(lineage_type),
                self.to_column().get_lineage(lineage_type),
            ],
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id}, from: {self.from_table()!r}, to: {self.to_table()!r})"



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class LabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    fontSize: Expression | None = None
    funnelLabelStyle: Expression | None = None
    labelDisplayUnits: Expression | None = None
    percentageLabelPrecision: Expression | None = None
    show: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper


class PercentBarLabelPropertiesHelper(LayoutNode):
    color: Expression | None = None
    show: Expression | None = None


class PercentBarLabelProperties(LayoutNode):
    properties: PercentBarLabelPropertiesHelper


class CategoryAxisPropertiesHelper(LayoutNode):
    color: Expression | None = None
    show: Expression | None = None


class CategoryAxisProperties(LayoutNode):
    properties: CategoryAxisPropertiesHelper


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class FunnelProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    labels: list[LabelsProperties] | None = None
    percentBarLabel: list[PercentBarLabelProperties] | None = None


class Funnel(BaseVisual):
    visualType: str = "funnel"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: FunnelProperties | None = None



import datetime
from typing import TYPE_CHECKING

from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.model_tables.enums import DataState
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.calculation_group import CalculationGroup
    from pbi_core.ssas.model_tables.format_string_definition import FormatStringDefinition
from pydantic import PrivateAttr


class CalculationItem(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/f5a398a7-ff65-45f0-a865-b561416f1cb4)
    """

    calculation_group_id: int
    description: str
    error_message: str
    expression: str
    format_string_definition_id: int
    name: str
    ordinal: int
    state: DataState

    modified_time: datetime.datetime
    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.calculation_item)

    def format_string_definition(self) -> "FormatStringDefinition":
        return self.tabular_model.format_string_definitions.find(self.format_string_definition_id)

    def calculation_group(self) -> "CalculationGroup":
        return self.tabular_model.calculation_groups.find(self.calculation_group_id)



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class BackgroundPropertiesHelper(LayoutNode):
    show: Expression | None = None
    transparency: Expression | None = None


class BackgroundProperties(LayoutNode):
    properties: BackgroundPropertiesHelper


class TitlePropertiesHelper(LayoutNode):
    alignment: Expression | None = None
    fontColor: Expression | None = None
    fontSize: Expression | None = None
    show: Expression | None = None
    text: Expression | None = None


class TitleProperties(LayoutNode):
    properties: TitlePropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    altText: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class LegendPropertiesHelper(LayoutNode):
    fontSize: Expression | None = None
    labelColor: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showTitle: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper


class LabelsPropertiesHelper(LayoutNode):
    background: Expression | None = None
    color: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelStyle: Expression | None = None
    overflow: Expression | None = None
    percentageLabelPrecision: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class SlicesPropertiesHelper(LayoutNode):
    innerRadiusRatio: Expression | None = None


class SlicesProperties(LayoutNode):
    properties: SlicesPropertiesHelper


class DonutChartProperties(LayoutNode):
    background: list[BackgroundProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    general: list[GeneralProperties] | None = None
    labels: list[LabelsProperties] | None = None
    legend: list[LegendProperties] | None = None
    slices: list[SlicesProperties] | None = None
    title: list[TitleProperties] | None = None


class DonutChart(BaseVisual):
    visualType: str = "donutChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: DonutChartProperties | None = None



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class CategoryLabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    show: Expression | None = None


class CategoryLabelsProperties(LayoutNode):
    properties: CategoryLabelsPropertiesHelper
    selector: Selector | None = None


class LabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    labelPrecision: Expression | None = None
    labelDisplayUnits: Expression | None = None
    preserveWhitespace: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper
    selector: Selector | None = None


class GeneralPropertiesHelper(LayoutNode):
    pass


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class WordWrapperPropertiesHelper(LayoutNode):
    show: Expression | None = None


class WordWrapProperties(LayoutNode):
    properties: WordWrapperPropertiesHelper


class CardProperties(LayoutNode):
    categoryLabels: list[CategoryLabelsProperties] | None = None
    general: list[GeneralProperties] | None = None
    labels: list[LabelsProperties] | None = None
    wordWrap: list[WordWrapProperties] | None = None


class Card(BaseVisual):
    visualType: str = "card"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: CardProperties | None = None



from enum import IntEnum


class ServerStateDiscoverBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class: 1: DISCOVER_CONNECTIONS 2: DISCOVER_SESSIONS 3: DISCOVER_TRANSACTIONS 6: DISCOVER_DB_CONNECTIONS 7: DISCOVER_JOBS 8: DISCOVER_LOCKS 12: DISCOVER_PERFORMANCE_COUNTERS 13: DISCOVER_MEMORYUSAGE 14: DISCOVER_JOB_PROGRESS 15: DISCOVER_MEMORYGRANT
    CURRENTTIME = 2  # Contains the current time of the server state discover event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the server state discover event.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the process ID of the client application that created the connection to the server.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the session ID associated with the server state discover event.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the server state discover event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the event.
    SERVERNAME = 43  # Contains the name of the instance on which the server state discover event occurred.
    REQUESTPROPERTIES = 45  # Contains the properties of the current XMLA request.
    ACTIVITYID = 46
    REQUESTID = 47


class ServerStateDiscoverDataColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class: 1: DISCOVER_CONNECTIONS 2: DISCOVER_SESSIONS 3: DISCOVER_TRANSACTIONS 6: DISCOVER_DB_CONNECTIONS 7: DISCOVER_JOBS 8: DISCOVER_LOCKS 12: DISCOVER_PERFORMANCE_COUNTERS 13: DISCOVER_MEMORYUSAGE 14: DISCOVER_JOB_PROGRESS 15: DISCOVER_MEMORYGRANT
    CURRENTTIME = 2  # Contains the current time of the server state discover event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the server state discover event.
    SESSIONID = 39  # Contains the session ID associated with the server state discover event.
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the server state discover event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with server response to the discover request.
    SERVERNAME = 43  # Contains the name of the instance on which the server state discover event occurred.


class ServerStateDiscoverEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class: 1: DISCOVER_CONNECTIONS 2: DISCOVER_SESSIONS 3: DISCOVER_TRANSACTIONS 6: DISCOVER_DB_CONNECTIONS 7: DISCOVER_JOBS 8: DISCOVER_LOCKS 12: DISCOVER_PERFORMANCE_COUNTERS 13: DISCOVER_MEMORYUSAGE 14: DISCOVER_JOB_PROGRESS 15: DISCOVER_MEMORYGRANT
    CURRENTTIME = 2  # Contains the current time of the server state discover event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Contains the time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Contains the amount of time (in milliseconds) taken by the event.
    CPUTIME = 6  # Contains the amount of CPU time (in milliseconds) used by the server state discover event.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the server state discover event.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the process ID of the client application that initiated the XMLA request.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the Windows domain account associated with the server state discover event.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the server state discover event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with server response to the discover request.
    SERVERNAME = 43  # Contains the name of the instance on which the server state discover event occurred.
    ACTIVITYID = 46
    REQUESTID = 47



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .properties.base import Expression


class FillPropertiesHelper(LayoutNode):
    fillColor: Expression | None = None
    image: Expression | None = None
    show: Expression | None = None
    transparency: Expression | None = None


class FillProperties(LayoutNode):
    properties: FillPropertiesHelper
    selector: Selector | None = None


class IconPropertiesHelper(LayoutNode):
    bottomMargin: Expression | None = None
    horizontalAlignment: Expression | None = None
    leftMargin: Expression | None = None
    lineColor: Expression | None = None
    lineTransparency: Expression | None = None
    lineWeight: Expression | None = None
    padding: Expression | None = None
    rightMargin: Expression | None = None
    shapeType: Expression | None = None
    show: Expression | None = None
    topMargin: Expression | None = None
    verticalAlignment: Expression | None = None


class IconProperties(LayoutNode):
    properties: IconPropertiesHelper
    selector: Selector | None = None


class TextPropertiesHelper(LayoutNode):
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    horizontalAlignment: Expression | None = None
    leftMargin: Expression | None = None
    padding: Expression | None = None
    rightMargin: Expression | None = None
    show: Expression | None = None
    text: Expression | None = None
    topMargin: Expression | None = None
    verticalAlignment: Expression | None = None


class TextProperties(LayoutNode):
    properties: TextPropertiesHelper
    selector: Selector | None = None


class OutlinePropertiesHelper(LayoutNode):
    lineColor: Expression | None = None
    roundEdge: Expression | None = None
    show: Expression | None = None
    transparency: Expression | None = None
    weight: Expression | None = None


class OutlineProperties(LayoutNode):
    properties: OutlinePropertiesHelper
    selector: Selector | None = None


class ShapePropertiesHelper(LayoutNode):
    roundEdge: Expression | None = None


class ShapeProperties(LayoutNode):
    properties: ShapePropertiesHelper
    selector: Selector | None = None


class ActionButtonProperties(LayoutNode):
    fill: list[FillProperties] | None = None
    icon: list[IconProperties] | None = None
    outline: list[OutlineProperties] | None = None
    shape: list[ShapeProperties] | None = None
    text: list[TextProperties] | None = None


class ActionButton(BaseVisual):
    visualType: str = "actionButton"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: ActionButtonProperties | None = None



from pathlib import Path
from typing import Any

import jinja2

from pbi_core.logging import get_logger
from pbi_core.pydantic import BaseValidation

logger = get_logger()

PACKAGE_DIR = Path(__file__).parents[2]
assert PACKAGE_DIR.name == "pbi_core"


class PbiCoreStartupConfig(BaseValidation):
    cert_dir: Path
    msmdsrv_ini: Path
    msmdsrv_exe: Path
    workspace_dir: Path

    def model_post_init(self, __context: Any, /) -> None:
        if not self.cert_dir.is_absolute():
            self.cert_dir = PACKAGE_DIR / self.cert_dir
        if not self.msmdsrv_ini.is_absolute():
            self.msmdsrv_ini = PACKAGE_DIR / self.msmdsrv_ini
        if not self.msmdsrv_exe.is_absolute():
            self.msmdsrv_exe = PACKAGE_DIR / self.msmdsrv_exe
        if not self.workspace_dir.is_absolute():
            self.workspace_dir = PACKAGE_DIR / self.workspace_dir

    def msmdsrv_ini_template(self) -> jinja2.Template:
        return jinja2.Template(self.msmdsrv_ini.read_text())


def get_startup_config() -> PbiCoreStartupConfig:
    try:
        with (PACKAGE_DIR / "local" / "settings.json").open("r") as f:
            return PbiCoreStartupConfig.model_validate_json(f.read())
    except FileNotFoundError as e:
        msg = 'Please run "python -m pbi_core setup" to initialize dependencies'
        raise FileNotFoundError(msg) from e



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    fillRule: Expression | None = None
    legend: Expression | None = None
    showAllDataPoints: Expression | None = None
    valueAxis: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class ValueAxisPropertiesHelper(LayoutNode):
    alignZeros: Expression | None = None
    axisScale: Expression | None = None
    end: Expression | None = None
    fontSize: Expression | None = None
    gridlineColor: Expression | None = None
    gridlineShow: Expression | None = None
    labelColor: Expression | None = None
    logAxisScale: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    switchAxisPosition: Expression | None = None
    titleColor: Expression | None = None
    titleFontFamily: Expression | None = None
    titleFontSize: Expression | None = None
    titleText: Expression | None = None
    treatNullsAsZero: Expression | None = None


class ValueAxisProperties(LayoutNode):
    properties: ValueAxisPropertiesHelper


class LegendPropertiesHelper(LayoutNode):
    fontSize: Expression | None = None
    labelColor: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showGradientLegend: Expression | None = None
    showTitle: Expression | None = None
    titleText: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper


class FillPointPropertiesHelper(LayoutNode):
    show: Expression | None = None
    style: Expression | None = None


class FillPointProperties(LayoutNode):
    properties: FillPointPropertiesHelper


class ColorBorderPropertiesHelper(LayoutNode):
    show: Expression | None = None


class ColorBorderProperties(LayoutNode):
    properties: ColorBorderPropertiesHelper


class CategoryAxisPropertiesHelper(LayoutNode):
    axisScale: Expression | None = None
    end: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    gridlineColor: Expression | None = None
    gridlineShow: Expression | None = None
    gridlineStyle: Expression | None = None
    innerPadding: Expression | None = None
    labelColor: Expression | None = None
    logAxisScale: Expression | None = None
    maxMarginFactor: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    titleColor: Expression | None = None
    titleFontFamily: Expression | None = None
    titleFontSize: Expression | None = None
    titleText: Expression | None = None
    treatNullsAsZero: Expression | None = None


class CategoryAxisProperties(LayoutNode):
    properties: CategoryAxisPropertiesHelper


class BubblesPropertiesHelper(LayoutNode):
    bubbleSize: Expression | None = None
    markerShape: Expression | None = None
    showSeries: Expression | None = None


class BubblesProperties(LayoutNode):
    properties: BubblesPropertiesHelper
    selector: Selector | None = None


class Y1AxisReferenceLinePropertiesHelper(LayoutNode):
    displayName: Expression | None = None
    lineColor: Expression | None = None
    show: Expression | None = None
    value: Expression | None = None


class Y1AxisReferenceLineProperties(LayoutNode):
    properties: Y1AxisReferenceLinePropertiesHelper
    selector: Selector | None = None


class CategoryLabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    enableBackground: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    show: Expression | None = None


class CategoryLabelsProperties(LayoutNode):
    properties: CategoryLabelsPropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    responsive: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class PlotAreaPropertiesHelper(LayoutNode):
    transparency: Expression | None = None


class PlotAreaProperties(LayoutNode):
    properties: PlotAreaPropertiesHelper


class ScatterChartProperties(LayoutNode):
    bubbles: list[BubblesProperties] | None = None
    categoryAxis: list[CategoryAxisProperties] | None = None
    categoryLabels: list[CategoryLabelsProperties] | None = None
    colorBorder: list[ColorBorderProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    fillPoint: list[FillPointProperties] | None = None
    general: list[GeneralProperties] | None = None
    legend: list[LegendProperties] | None = None
    plotArea: list[PlotAreaProperties] | None = None
    valueAxis: list[ValueAxisProperties] | None = None
    y1AxisReferenceLine: list[Y1AxisReferenceLineProperties] | None = None


class ScatterChart(BaseVisual):
    visualType: str = "scatterChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: ScatterChartProperties | None = None



from enum import IntEnum


class ProgressReportBeginColumns(IntEnum):
    pass


class ProgressReportEndColumns(IntEnum):
    pass


class ProgressReportCurrentColumns(IntEnum):
    pass


class ProgressReportErrorColumns(IntEnum):
    pass



from typing import Annotated, Any

from pydantic import Discriminator, Tag

from .action_button import ActionButton
from .bar_chart import BarChart
from .basic_shape import BasicShape
from .card import Card
from .clustered_column_chart import ClusteredColumnChart
from .column_chart import ColumnChart
from .donut_chart import DonutChart
from .funnel import Funnel
from .generic import GenericVisual
from .image import Image
from .line_chart import LineChart
from .line_stacked_column_combo_chart import LineStackedColumnComboChart
from .pie_chart import PieChart
from .scatter_chart import ScatterChart
from .slicer import Slicer
from .table import TableChart
from .text_box import TextBox


def get_visual(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        assert "visualType" in v
        assert isinstance(v["visualType"], str)

        mapping = {
            "actionButton": "ActionButton",
            "barChart": "BarChart",
            "basicShape": "BasicShape",
            "card": "Card",
            "clusteredColumnChart": "ClusteredColumnChart",
            "columnChart": "ColumnChart",
            "donutChart": "DonutChart",
            "funnel": "Funnel",
            "image": "Image",
            "lineChart": "LineChart",
            "lineStackedColumnComboChart": "LineStackedColumnComboChart",
            "pieChart": "PieChart",
            "scatterChart": "ScatterChart",
            "slicer": "Slicer",
            "tableEx": "TableChart",
            "textbox": "TextBox",
        }
        return mapping.get(v["visualType"], "GenericVisual")
    return v.__class__.__name__


Visual = Annotated[
    Annotated[ActionButton, Tag("ActionButton")]
    | Annotated[BarChart, Tag("BarChart")]
    | Annotated[GenericVisual, Tag("GenericVisual")]
    | Annotated[BasicShape, Tag("BasicShape")]
    | Annotated[Card, Tag("Card")]
    | Annotated[ColumnChart, Tag("ColumnChart")]
    | Annotated[ClusteredColumnChart, Tag("ClusteredColumnChart")]
    | Annotated[DonutChart, Tag("DonutChart")]
    | Annotated[Funnel, Tag("Funnel")]
    | Annotated[Image, Tag("Image")]
    | Annotated[LineChart, Tag("LineChart")]
    | Annotated[LineStackedColumnComboChart, Tag("LineStackedColumnComboChart")]
    | Annotated[PieChart, Tag("PieChart")]
    | Annotated[ScatterChart, Tag("ScatterChart")]
    | Annotated[Slicer, Tag("Slicer")]
    | Annotated[TableChart, Tag("TableChart")]
    | Annotated[TextBox, Tag("TextBox")],
    Discriminator(get_visual),
]


__all__ = ["GenericVisual", "Visual"]



from enum import Enum, IntEnum


class DataState(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/93d1844f-a6c7-4dda-879b-2e26ed5cd297)."""

    READY = 1
    NO_DATA = 3
    CALCULATION_NEEDED = 4
    SEMANTIC_ERROR = 5
    EVALUATION_ERROR = 6
    DEPENDENCY_ERROR = 7
    INCOMPLETE = 8
    SYNTAX_ERROR = 9


class ObjectType(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/7a16a837-cb88-4cb2-a766-a97c4d0e1f43)."""

    MODEL = 1
    DATASOURCE = 2
    TABLE = 3
    COLUMN = 4
    ATTRIBUTE_HIERARCHY = 5
    PARTITION = 6
    RELATIONSHIP = 7
    MEASURE = 8
    HIERARCHY = 9
    LEVEL = 10
    KPI = 12
    CULTURE = 13
    LINGUISTIC_METADATA = 15
    PERSPECTIVE = 29
    PERSPECTIVE_TABLE = 30
    PERSPECTIVE_COLUMN = 31
    PERSPECTIVE_HIERARCHY = 32
    PERSPECTIVE_MEASURE = 33
    ROLE = 34
    ROLE_MEMBERSHIP = 35
    TABLE_PERMISSION = 36
    VARIATION = 37
    EXPRESSION = 41
    COLUMN_PERMISSION = 42
    CALCULATION_GROUP = 46
    CALCULATION_ITEM = 47
    QUERY_GROUP = 51


class DataType(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/00a9ec7a-5f4d-4517-8091-b370fe2dc18b)."""

    AUTOMATIC = 1
    STRING = 2
    INT64 = 6
    DOUBLE = 8
    DATE_TIME = 9
    DECIMAL = 10
    BOOLEAN = 11
    BINARY = 17
    UNKNOWN = 19
    UNKNOWNER = 20


class DataCategory(Enum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/6360ac84-0717-4170-bce0-284cbef419ca).

    Note:
        Only used for table, the Column and Measure DataCategories are just strings

    """

    UNKNOWN = 0
    REGULAR = 1
    TIME = 2
    GEOGRAPHY = 3
    ORGANIZATION = 4
    BILL_OF_MATERIALS = 5
    ACCOUNTS = 6
    CUSTOMERS = 7
    PRODUCTS = 8
    SCENARIO = 9
    QUANTITATIVE = 10
    UTILITY = 11
    CURRENCY = 12
    RATES = 13
    CHANNEL = 14
    PROMOTION = 15
    TIME_STR = "Time"  # TODO: I think we're accidentally merging two enums. sometimes it's a string in the pbix??



from pbi_core.static_files.layout._base_node import LayoutNode

from .base import SourceExpression


class MeasureSource(LayoutNode):
    Measure: SourceExpression
    Name: str | None = None
    NativeReferenceName: str | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MeasureSource):
            return False
        return (self.Measure.column() == other.Measure.column()) and (self.Measure.table() == other.Measure.table())

    def __hash__(self) -> int:
        return hash((self.Measure.table(), self.Measure.column()))



import datetime
from typing import TYPE_CHECKING, Literal
from uuid import UUID, uuid4

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.hierarchy import Hierarchy


class Level(SsasRenameRecord):
    """A level in a hierarchy. For example, in a hierarchy of "Date", the levels could be "Year", "Month", and "Day".

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/a010d75e-3b68-4825-898f-62fdeab4557f)
    """

    column_id: int
    description: str | None = None
    """A description of the level, which may be used in the hover tooltip in edit mode"""
    hierarchy_id: int
    name: str
    """The name of the level, e.g. "Year", "Quarter", "Month", "Day" in a Date hierarchy."""
    ordinal: int

    lineage_tag: UUID = uuid4()
    source_lineage_tag: UUID = uuid4()

    modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.level)

    def column(self) -> "Column":
        return self.tabular_model.columns.find({"id": self.column_id})

    def hierarchy(self) -> "Hierarchy":
        return self.tabular_model.hierarchies.find({"id": self.hierarchy_id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        return LineageNode(
            self,
            lineage_type,
            [self.column().get_lineage(lineage_type), self.hierarchy().get_lineage(lineage_type)],
        )



from pbi_core.ssas.server.tabular_model.tabular_model import LocalTabularModel

from .performance_trace import Performance, PerformanceTrace


def get_performance(model: LocalTabularModel, commands: list[str], *, clear_cache: bool = False) -> list[Performance]:
    """Calculates performance of a DAX query using a Trace.

    Args:
        model (LocalTabularModel): the SSAS instance the DAX should be run against
        commands (list[str]): A list of DAX queries to run against the SSAS model
        clear_cache (bool): Whether to clear the SSAS cache before running the queries.
            Useful to test cold start times for users

    Returns:
        Performance: contains memory and time usage of the DAX query


    """
    perf_trace = PerformanceTrace(model, commands)
    return perf_trace.get_performance(clear_cache=clear_cache)



from typing import TYPE_CHECKING

import jinja2

from pbi_core.ssas.server._commands import BASE_ALTER_TEMPLATE

if TYPE_CHECKING:
    from .ssas_tables import SsasAlter


BATCH_COMMAND = jinja2.Template("""
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
    {{inner_commands}}
</Batch>
""")


class AlterCommand:
    database_id: str
    object_types: dict[str, list["SsasAlter"]]

    def __init__(self, database_id: str, object_types: dict[str, list["SsasAlter"]]) -> None:
        self.database_id = database_id
        self.object_types = object_types

    def render_xml(self) -> str:
        entity_commands: list[str] = []
        for entities in self.object_types.values():
            entity_rows = [
                table._get_row_xml(
                    table.xml_fields(),
                    table._commands.alter,
                )
                for table in entities
            ]
            entity_rows_str = "\n".join(entity_rows)
            entity_command = entities[0]._commands.alter.entity_template.render(rows=entity_rows_str)
            entity_commands.append(entity_command)
        ret = BASE_ALTER_TEMPLATE.render(
            db_name=self.database_id,
            entity_def="\n".join(entity_commands),
        )
        return "\n".join(ret.splitlines()[1:-1])  # the first and last are the batch start/end


class Batch:
    commands: list[AlterCommand]

    def __init__(self, commands: list[AlterCommand]) -> None:
        self.commands = commands

    def render_xml(self) -> str:
        command_strs = [command.render_xml() for command in self.commands]
        return BATCH_COMMAND.render(inner_commands="\n".join(command_strs))



import logging
from typing import cast

import structlog


def get_logger(log_level: int = logging.INFO) -> structlog.BoundLogger:
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
    )
    return cast("structlog.BoundLogger", structlog.get_logger())



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.expansion_state import ExpansionState
from pbi_core.static_files.layout.filters import Filter
from pbi_core.static_files.layout.selector import SelectorData

from .base import BaseVisual
from .properties.base import Expression
from .table import ColumnProperty


class SyncGroup(LayoutNode):
    groupName: str
    fieldChanges: bool
    filterChanges: bool = True


class CachedFilterDisplayItems(LayoutNode):
    id: SelectorData
    displayName: str


class HeaderPropertiesHelper(LayoutNode):
    background: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    show: Expression | None = None
    showRestatement: Expression | None = None
    textSize: Expression | None = None


class HeaderProperties(LayoutNode):
    properties: HeaderPropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    filter: Filter | None = None
    responsive: Expression | None = None
    selfFilterEnabled: Expression | None = None
    selfFilter: Filter | None = None
    orientation: Expression | None = None
    outlineColor: Expression | None = None
    outlineWeight: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class DataPropertiesHelper(LayoutNode):
    endDate: Expression | None = None
    isInvertedSelectionMode: Expression | None = None
    mode: Expression | None = None
    numericEnd: Expression | None = None
    numericStart: Expression | None = None
    startDate: Expression | None = None


class DataProperties(LayoutNode):
    properties: DataPropertiesHelper


class ItemPropertiesHelper(LayoutNode):
    background: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    outline: Expression | None = None
    outlineColor: Expression | None = None
    textSize: Expression | None = None


class ItemProperties(LayoutNode):
    properties: ItemPropertiesHelper


class DatePropertiesHelper(LayoutNode):
    background: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    textSize: Expression | None = None


class DateProperties(LayoutNode):
    properties: DatePropertiesHelper


class SliderPropertiesHelper(LayoutNode):
    color: Expression | None = None
    show: Expression | None = None


class SliderProperties(LayoutNode):
    properties: SliderPropertiesHelper


class SelectionPropertiesHelper(LayoutNode):
    selectAllCheckboxEnabled: Expression | None = None
    singleSelect: Expression | None = None
    strictSingleSelect: Expression | None = None


class SelectionProperties(LayoutNode):
    properties: SelectionPropertiesHelper


class NumericInputStylePropertiesHelper(LayoutNode):
    background: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    textSize: Expression | None = None


class NumericInputStyleProperties(LayoutNode):
    properties: NumericInputStylePropertiesHelper


class SlicerProperties(LayoutNode):
    date: list[DateProperties] | None = None
    data: list[DataProperties] | None = None
    general: list[GeneralProperties] | None = None
    header: list[HeaderProperties] | None = None
    items: list[ItemProperties] | None = None
    numericInputStyle: list[NumericInputStyleProperties] | None = None
    selection: list[SelectionProperties] | None = None
    slider: list[SliderProperties] | None = None


class Slicer(BaseVisual):
    visualType: str = "slicer"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    syncGroup: SyncGroup | None = None
    cachedFilterDisplayItems: list[CachedFilterDisplayItems] | None = None
    expansionStates: list[ExpansionState] | None = None
    objects: SlicerProperties | None = None



import json
import shutil
from pathlib import Path

from pbi_core.ssas.setup import PbiCoreStartupConfig

settings_path = Path(__file__).parents[1] / "local" / "settings.json"


def get_workspaces() -> list[Path]:
    settings = PbiCoreStartupConfig.model_validate(json.load(settings_path.open("r")))
    return list(settings.workspace_dir.iterdir())


def list_workspaces() -> None:
    settings = PbiCoreStartupConfig.model_validate(json.load(settings_path.open("r")))
    workspaces = get_workspaces()
    if len(workspaces) == 0:
        print(f"No workspaces to remove from {settings.workspace_dir.absolute().as_posix()}")
    else:
        print("Folders to remove:")
        for folder in workspaces:
            print(folder)


def clear_workspaces() -> None:
    settings = PbiCoreStartupConfig.model_validate(json.load(settings_path.open("r")))
    workspaces = get_workspaces()
    if len(workspaces) == 0:
        print(f"No workspaces to remove from {settings.workspace_dir.absolute().as_posix()}")

    for folder in workspaces:
        print(f"Deleting: {folder.absolute().as_posix()}")
        try:
            shutil.rmtree(folder)
        except PermissionError:
            print("\tWorkspace currently being used by an SSAS instance")



from enum import IntEnum


class ErrorColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    SESSIONTYPE = 8  # Contains the type of the entity that caused the error.
    SEVERITY = 22  # Contains the severity level of an exception associated with the error event. Values are: 0 = Success 1 = Informational 2 = Warning 3 = Error
    SUCCESS = 23  # Contains the success or failure of the error event. Values are: 0 = Failure 1 = Success
    ERROR = 24  # Contains the error number of any error associated with the error event.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the error event.
    DATABASENAME = 28  # Contains the name of the Analysis Services instance on which the error event occurred.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Contains the name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # Contains the process ID of the client application.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the error event. The SPID directly corresponds to the session GUID used by XML for Analysis (XMLA).
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the error event. The SPID directly corresponds to the session GUID used by XML for Analysis (XMLA).
    TEXTDATA = 42  # Contains the text data associated with the error event.
    SERVERNAME = (
        43  # Contains the name of the server running Analysis Services instance on which the error event occurred.
    )
    ACTIVITYID = 46
    REQUESTID = 47



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode

from ..selector import Selector
from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class LegendPropertiesHelper(LayoutNode):
    position: Expression | None = None
    show: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class LabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelPrecision: Expression | None = None
    labelStyle: Expression | None = None
    percentageLabelPrecision: Expression | None = None
    show: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper


class PieChartProperties(LayoutNode):
    dataPoint: list[DataPointProperties] | None = None
    labels: list[LabelsProperties] | None = None
    legend: list[LegendProperties] | None = None


class PieChart(BaseVisual):
    visualType: str = "pieChart"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    objects: PieChartProperties | None = None



import os
import shutil
from pathlib import Path

import bs4
import inquirer

from .config import PACKAGE_DIR, PbiCoreStartupConfig


def get_msmdsrv_exe_path() -> str | None:
    """Can require admin rights to do."""
    candidate_folders = ["C:/Program Files/Microsoft Power BI Desktop", "C:/Program Files/WindowsApps"]
    for folder in candidate_folders:
        for path in Path(folder).glob("**/msmdsrv.exe"):
            return path.parent.absolute().as_posix()
    return None


def get_msmdsrv_ini_path() -> str | None:
    """Can require admin rights to do."""
    candidate_folders = [
        f"C:/Users/{os.getlogin()}/Microsoft/Power BI Desktop Store App/AnalysisServicesWorkspaces",
    ]
    for folder in candidate_folders:
        for path in Path(folder).glob("**/msmdsrv.ini"):
            return path.absolute().as_posix()
    return None


def get_certificate_directory() -> str | None:
    candidates = [
        f"C:/Users/{os.getlogin()}/AppData/Local/Microsoft/Power BI Desktop/CertifiedExtensions",
        f"C:/Users/{os.getlogin()}/Microsoft/Power BI Desktop Store App/CertifiedExtensions",
    ]
    for candidate in candidates:
        if Path(candidate).exists():
            return candidate
    return None


def validator(_: dict, inp: str) -> bool:
    return Path(inp).exists()


def validator_potential(_: dict, inp: str) -> bool:
    try:
        Path(inp).mkdir(exist_ok=True, parents=True)
    except Exception:  # noqa: BLE001
        return False
    else:
        return True


def create_ini_template(raw_path: Path) -> None:
    ini_contents = bs4.BeautifulSoup(raw_path.open(encoding="utf-8"), "xml")
    for tag in ["DataDir", "TempDir", "LogDir", "BackupDir", "CrashReportsFolder", "AllowedBrowsingFolders"]:
        ini_contents.find(tag).string = "{{data_directory}}"  # pyright: ignore reportAttributeAccessIssue
    ini_contents.find("CertifiedExtensions").Directories.string = "{{certificate_directory}}"  # pyright: ignore reportAttributeAccessIssue
    ini_contents.find("ExtensionDir").string = "{{bin_directory}}"  # pyright: ignore reportAttributeAccessIssue
    ini_contents.find("DataSourceImpersonationAccount").Password.string = ""  # pyright: ignore reportAttributeAccessIssue
    ini_contents.find("QueryLogConnectionString").string = ""  # pyright: ignore reportAttributeAccessIssue
    ini_contents.find(
        "PrivateProcess",
    ).string = "4"  # pyright: ignore reportAttributeAccessIssue  # this must point to an existing process (I think??). 4 is the system, so shouldn't ever fail
    with Path(PACKAGE_DIR / "local" / "msmdsrv_template.ini").open("w", encoding="utf-8") as f:
        # [39:] to remove "<?xml version="1.0" encoding="utf-8"?>"
        f.write(str(ini_contents)[39:])


def interactive_setup() -> None:
    cert_dir = get_certificate_directory()
    msmdsrv_exe_path = get_msmdsrv_exe_path()
    msmdsrv_ini_path = get_msmdsrv_ini_path()

    target_dir = PACKAGE_DIR / "local"

    default_workspace_path = (PACKAGE_DIR / "workspaces").absolute().as_posix()
    questions = [
        inquirer.Text(
            "workspace_dir",
            message="Path for temp workspaces",
            default=default_workspace_path,
            validate=validator_potential,
        ),
        inquirer.Text(
            "msmdsrv_ini",
            message="Path to PowerBI's msmdsrv.ini",
            default=msmdsrv_ini_path,
            validate=validator,
        ),
        inquirer.Text(
            "bin_path",
            message="Path to PowerBI's bin folder",
            default=msmdsrv_exe_path,
            validate=validator,
        ),
        inquirer.Text(
            "cert_dir",
            message="Path to PowerBI's CertifiedExtensions folder",
            default=cert_dir,
            validate=validator,
        ),
    ]
    answers = inquirer.prompt(questions)
    assert answers is not None

    target_dir.mkdir(exist_ok=True)
    shutil.rmtree(target_dir)
    (target_dir / "pbi").mkdir(exist_ok=True, parents=True)

    shutil.copy(answers["msmdsrv_ini"], target_dir / "msmdsrv.ini")
    create_ini_template(target_dir / "msmdsrv.ini")

    shutil.copytree(answers["bin_path"], target_dir / "bin")
    shutil.copytree(answers["cert_dir"], target_dir / "CertifiedExtensions")

    config_data = {
        "msmdsrv_ini": target_dir / "msmdsrv_template.ini",
        "msmdsrv_exe": target_dir / "bin" / "msmdsrv.exe",
        "cert_dir": target_dir / "CertifiedExtensions",
        "workspace_dir": Path(answers["workspace_dir"]),
    }
    for k, v in config_data.items():
        if PACKAGE_DIR in v.parents:
            config_data[k] = v.relative_to(PACKAGE_DIR)

    out = PbiCoreStartupConfig(**config_data).model_dump_json(indent=2)
    with (target_dir / "settings.json").open("w") as f:
        f.write(out)



from enum import IntEnum
from typing import TYPE_CHECKING
from uuid import UUID

from pydantic import Json

from ._base_node import LayoutNode
from .sources import ColumnSource

if TYPE_CHECKING:
    from .layout import Layout


class Parameter(LayoutNode):
    _parent: "Pod"  # pyright: ignore reportIncompatibleVariableOverride=false

    name: str
    boundFilter: str
    fieldExpr: ColumnSource | None = None
    isLegacySingleSelection: bool | None = False
    asAggregation: bool | None = False


class PodType(IntEnum):
    NA1 = 1
    NA2 = 2


class PodConfig(LayoutNode):
    acceptsFilterContext: bool = False


class Pod(LayoutNode):
    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false

    id: int | None = None
    name: str
    boundSection: str
    config: Json[PodConfig]
    parameters: Json[list[Parameter]] = []
    type: PodType | None = None
    referenceScope: int | None = None
    cortanaEnabled: bool | None = None
    objectId: UUID | None = None



from typing import cast

from pbi_core.static_files.layout._base_node import LayoutNode

from .aggregation import DataSource
from .base import SourceRef
from .column import ColumnSource


class _GroupSourceHelper(LayoutNode):
    _parent: "GroupSource"  # pyright: ignore reportIncompatibleVariableOverride=false

    Expression: SourceRef
    GroupedColumns: list[ColumnSource]
    Property: str


class GroupSource(LayoutNode):
    GroupRef: _GroupSourceHelper
    Name: str | None = None

    def __repr__(self) -> str:
        table = self.GroupRef.Expression.table()
        column = self.GroupRef.Property
        return f"GroupRef({table}.{column})"

    def filter_name(self) -> str:
        return self.GroupRef.Property

    def get_sources(self) -> list[DataSource]:
        return cast("list[DataSource]", self.GroupRef.GroupedColumns)



import datetime
from typing import TYPE_CHECKING

from pbi_parsers import dax
from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord, SsasTable
from pbi_core.ssas.model_tables.enums import DataState, ObjectType
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables._group import Group


class FormatStringDefinition(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/b756b0c1-c912-4218-80dc-7ff09d0968ff)
    """

    object_type: ObjectType
    object_id: int
    error_message: str | None = None
    """When no issue exists, this field is blank"""
    expression: str
    """The DAX expression defining the """
    state: DataState

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.format_string_definition)

    def pbi_core_name(self) -> str:
        return str(self.object_id)

    def object(self) -> SsasTable:
        """Returns the object the annotation is describing.

        Raises:
            TypeError: When the Object Type doesn't map to a know SSAS entity type

        """
        if self.object_type == ObjectType.MODEL:
            return self.tabular_model.model

        type_mapper: dict[ObjectType, Group] = {
            ObjectType.DATASOURCE: self.tabular_model.data_sources,
            ObjectType.TABLE: self.tabular_model.tables,
            ObjectType.COLUMN: self.tabular_model.columns,
            ObjectType.ATTRIBUTE_HIERARCHY: self.tabular_model.attribute_hierarchies,
            ObjectType.PARTITION: self.tabular_model.partitions,
            ObjectType.RELATIONSHIP: self.tabular_model.relationships,
            ObjectType.MEASURE: self.tabular_model.measures,
            ObjectType.HIERARCHY: self.tabular_model.hierarchies,
            ObjectType.LEVEL: self.tabular_model.levels,
            ObjectType.KPI: self.tabular_model.kpis,
            ObjectType.CULTURE: self.tabular_model.cultures,
            ObjectType.LINGUISTIC_METADATA: self.tabular_model.linguistic_metadata,
            ObjectType.PERSPECTIVE: self.tabular_model.perspectives,
            ObjectType.PERSPECTIVE_TABLE: self.tabular_model.perspective_tables,
            ObjectType.PERSPECTIVE_HIERARCHY: self.tabular_model.perspective_hierarchies,
            ObjectType.PERSPECTIVE_MEASURE: self.tabular_model.perspective_measures,
            ObjectType.ROLE: self.tabular_model.roles,
            ObjectType.ROLE_MEMBERSHIP: self.tabular_model.role_memberships,
            ObjectType.TABLE_PERMISSION: self.tabular_model.table_permissions,
            ObjectType.VARIATION: self.tabular_model.variations,
            ObjectType.EXPRESSION: self.tabular_model.expressions,
            ObjectType.COLUMN_PERMISSION: self.tabular_model.column_permissions,
            ObjectType.CALCULATION_GROUP: self.tabular_model.calculation_groups,
            ObjectType.QUERY_GROUP: self.tabular_model.query_groups,
        }
        if self.object_type not in type_mapper:
            msg = f"Object Type {self.object_type} does not map to a known SSAS entity type."
            raise TypeError(msg)

        return type_mapper[self.object_type].find(self.object_id)

    def expression_ast(self) -> dax.Expression | None:
        if not isinstance(self.expression, str):
            return None
        ret = dax.to_ast(self.expression)
        if ret is None:
            msg = "Failed to parse DAX expression from format string definition"
            raise ValueError(msg)
        return ret

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id}, on: {self.object()!r})"



import datetime
from typing import Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord, SsasTable
from pbi_core.ssas.model_tables.enums import ObjectType
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands


class Annotation(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/7a16a837-cb88-4cb2-a766-a97c4d0e1f43)
    """

    object_id: int
    object_type: ObjectType
    name: str
    value: str | None = None

    modified_time: datetime.datetime
    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.annotation)

    def object(self) -> SsasTable:
        """Returns the object the annotation is describing.

        Raises:
            TypeError: When the Object Type doesn't map to a know SSAS entity type

        """
        mapper = {
            ObjectType.ATTRIBUTE_HIERARCHY: self.tabular_model.attribute_hierarchies.find,
            ObjectType.CALCULATION_GROUP: self.tabular_model.calculation_groups.find,
            ObjectType.COLUMN: self.tabular_model.columns.find,
            ObjectType.COLUMN_PERMISSION: self.tabular_model.column_permissions.find,
            ObjectType.CULTURE: self.tabular_model.cultures.find,
            ObjectType.DATASOURCE: self.tabular_model.data_sources.find,
            ObjectType.EXPRESSION: self.tabular_model.expressions.find,
            ObjectType.HIERARCHY: self.tabular_model.hierarchies.find,
            ObjectType.KPI: self.tabular_model.kpis.find,
            ObjectType.LEVEL: self.tabular_model.levels.find,
            ObjectType.LINGUISTIC_METADATA: self.tabular_model.linguistic_metadata.find,
            ObjectType.MEASURE: self.tabular_model.measures.find,
            ObjectType.MODEL: lambda _x: self.tabular_model.model,
            ObjectType.PARTITION: self.tabular_model.partitions.find,
            ObjectType.PERSPECTIVE: self.tabular_model.perspectives.find,
            ObjectType.PERSPECTIVE_HIERARCHY: self.tabular_model.perspective_hierarchies.find,
            ObjectType.PERSPECTIVE_MEASURE: self.tabular_model.perspective_measures.find,
            ObjectType.PERSPECTIVE_TABLE: self.tabular_model.perspective_tables.find,
            ObjectType.QUERY_GROUP: self.tabular_model.query_groups.find,
            ObjectType.RELATIONSHIP: self.tabular_model.relationships.find,
            ObjectType.ROLE: self.tabular_model.roles.find,
            ObjectType.ROLE_MEMBERSHIP: self.tabular_model.role_memberships.find,
            ObjectType.TABLE: self.tabular_model.tables.find,
            ObjectType.TABLE_PERMISSION: self.tabular_model.table_permissions.find,
            ObjectType.VARIATION: self.tabular_model.variations.find,
        }
        if self.object_type not in mapper:
            msg = f"Object Type {self.object_type} does not map to a known SSAS entity type."
            raise TypeError(msg)
        return mapper[self.object_type](self.object_id)

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        return LineageNode(self, lineage_type, [self.object().get_lineage(lineage_type)])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id}, on: {self.object()!r}, name: {self.name}, value: {self.value})"



from uuid import UUID

from ._base import BaseFileModel


class RemoteArtifact(BaseFileModel):
    DatasetId: UUID
    ReportId: UUID


class Connections(BaseFileModel):
    Version: int
    RemoteArtifacts: list[RemoteArtifact]



from enum import IntEnum


class PartitionMode(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/analysis-services/tmsl/partitions-object-tmsl?view=asallproducts-allversions)."""

    IMPORT = 0
    DIRECT_QUERY = 1  # not verified
    DEFAULT = 2  # not verified
    PUSH = 3


class PartitionType(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/analysis-services/tmsl/partitions-object-tmsl?view=asallproducts-allversions)."""

    QUERY = 1
    CALCULATED = 2
    _NONE = 3
    M = 4
    ENTITY = 5
    CALCULATION_GROUP = 7


class DataView(IntEnum):
    FULL = 0
    SAMPLE = 1
    DEFAULT = 3



import json
import zipfile
from dataclasses import dataclass
from typing import TYPE_CHECKING
from zipfile import ZipFile

import bs4

from .file_classes import Connections, DiagramLayout, Metadata, Settings, parse_diagram_layout
from .file_classes.theme import Theme
from .layout._base_node import _set_parents
from .layout.layout import Layout

if TYPE_CHECKING:
    from _typeshed import StrPath

LAYOUT_ENCODING = "utf-16-le"


@dataclass
class Version:
    major: int
    minor: int

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"


class StaticFiles:
    content_types: bs4.BeautifulSoup
    connections: Connections | None
    # no datamodel, that's handled by the ssas folder
    diagram_layout: DiagramLayout
    layout: Layout
    metadata: Metadata
    version: Version
    security_bindings: bytes
    settings: Settings
    themes: dict[str, Theme]

    def __init__(  # noqa: PLR0913, PLR0917
        self,
        content_types: bs4.BeautifulSoup,
        connections: Connections | None,
        diagram_layout: DiagramLayout,
        layout: Layout,
        metadata: Metadata,
        version: Version,
        security_bindings: bytes,
        settings: Settings,
        themes: dict[str, Theme],
    ) -> None:
        self.content_types = content_types
        self.connections = connections
        self.diagram_layout = diagram_layout
        self.layout = layout
        self.metadata = metadata
        self.version = version
        self.security_bindings = security_bindings
        self.settings = settings
        self.themes = themes

    def __repr__(self) -> str:
        return f"StaticFiles(version={self.version})"

    @staticmethod
    def load_pbix(path: "StrPath") -> "StaticFiles":  # noqa: PLR0914
        zipfile = ZipFile(path, mode="r")

        themes: dict[str, Theme] = {}
        theme_paths = [
            x.split("/")[-1]
            for x in zipfile.namelist()
            if x.startswith("Report/StaticResources/SharedResources/BaseThemes")
        ]
        for theme_path in theme_paths:
            theme_json = json.loads(
                zipfile.read(f"Report/StaticResources/SharedResources/BaseThemes/{theme_path}").decode("utf-8"),
            )
            themes[theme_path] = Theme.model_validate(theme_json)

        layout_json = json.loads(zipfile.read("Report/Layout").decode(LAYOUT_ENCODING))
        layout = Layout.model_validate(layout_json)
        _set_parents(layout, None, [])  # pyright: ignore reportArgumentType

        if "Connections" in zipfile.namelist():
            connections_json = json.loads(zipfile.read("Connections").decode("utf-8"))
            connections = Connections.model_validate(connections_json)
        else:
            connections = None

        major, minor = zipfile.read("Version").decode(LAYOUT_ENCODING).split(".")
        version = Version(int(major), int(minor))

        security_bindings = zipfile.read("SecurityBindings")

        content_types = bs4.BeautifulSoup(zipfile.read("[Content_Types].xml").decode("utf-8"), "xml")

        metadata_json = json.loads(zipfile.read("Metadata").decode(LAYOUT_ENCODING))
        metadata = Metadata.model_validate(metadata_json)

        diagram_json = json.loads(zipfile.read("DiagramLayout").decode(LAYOUT_ENCODING))
        diagram_layout = parse_diagram_layout(diagram_json)

        settings_json = json.loads(zipfile.read("Settings").decode(LAYOUT_ENCODING))
        settings = Settings.model_validate(settings_json)
        return StaticFiles(
            content_types,
            connections,
            diagram_layout,
            layout,
            metadata,
            version,
            security_bindings,
            settings,
            themes,
        )

    def save_pbix(self, path: "StrPath") -> None:
        """Saves report to file.

        We use "a" as the write mode because the ssas.save_pbix method already
        creates a PBIX with the datamodel subfile.
        """
        data: dict[str, bytes] = {}
        with zipfile.ZipFile(path, "r", compression=zipfile.ZIP_DEFLATED) as zf:
            for f in zf.namelist():
                data[f] = zf.read(f)

        with zipfile.ZipFile(path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("Settings", self.settings.model_dump_json().encode(LAYOUT_ENCODING))
            if self.connections is not None:
                zf.writestr("Connections", self.connections.model_dump_json().encode("utf-8"))
            zf.writestr("Version", str(self.version).encode(LAYOUT_ENCODING))
            zf.writestr(
                "DiagramLayout",
                self.diagram_layout.model_dump_json().encode(LAYOUT_ENCODING),
            )  # has null nodeLineageTag that the source doesn't
            for path_name, theme_info in self.themes.items():
                zf.writestr(
                    f"Report/StaticResources/SharedResources/BaseThemes/{path_name}",
                    theme_info.model_dump_json().encode("utf-8"),
                )
            zf.writestr("DataModel", data["DataModel"])
            zf.writestr("Report/Layout", self.layout.model_dump_json().encode(LAYOUT_ENCODING))
            zf.writestr("[Content_Types].xml", str(self.content_types).encode("utf8"))



from typing import Annotated, Any

from pydantic import Discriminator, Tag

from pbi_core.pydantic.main import BaseValidation

from .aggregation import AggregationSource, DataSource, SelectRef
from .arithmetic import ArithmeticSource
from .base import Entity, SourceRef
from .column import ColumnSource
from .group import GroupSource
from .hierarchy import HierarchyLevelSource
from .literal import LiteralSource
from .measure import MeasureSource
from .proto import ProtoSourceRef


class RoleRef(BaseValidation):
    Role: str


class TransformOutputRoleRef(BaseValidation):
    TransformOutputRoleRef: RoleRef
    Name: str | None = None


def get_source(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        mapper = {
            "Column": "ColumnSource",
            "HierarchyLevel": "HierarchyLevelSource",
            "GroupRef": "GroupSource",
            "Aggregation": "AggregationSource",
            "Measure": "MeasureSource",
            "Arithmetic": "ArithmeticSource",
            "SourceRef": "ProtoSourceRef",
            "TransformOutputRoleRef": "TransformOutputRoleRef",
            "Literal": "LiteralSource",
            "SelectRef": "SelectRef",
        }
        for key in v:
            if key in mapper:
                return mapper[key]
        msg = f"Unknown Filter: {v.keys()}"
        raise TypeError(msg)
    return v.__class__.__name__


Source = Annotated[
    Annotated[HierarchyLevelSource, Tag("HierarchyLevelSource")]
    | Annotated[ColumnSource, Tag("ColumnSource")]
    | Annotated[GroupSource, Tag("GroupSource")]
    | Annotated[AggregationSource, Tag("AggregationSource")]
    | Annotated[MeasureSource, Tag("MeasureSource")]
    | Annotated[ArithmeticSource, Tag("ArithmeticSource")]
    | Annotated[ProtoSourceRef, Tag("ProtoSourceRef")]
    | Annotated[TransformOutputRoleRef, Tag("TransformOutputRoleRef")]
    | Annotated[LiteralSource, Tag("LiteralSource")]
    | Annotated[SelectRef, Tag("SelectRef")],
    # Discriminator is used to determine the type based on the content of the object
    Discriminator(get_source),
]

__all__ = [
    "AggregationSource",
    "ArithmeticSource",
    "ColumnSource",
    "DataSource",
    "Entity",
    "GroupSource",
    "HierarchyLevelSource",
    "LiteralSource",
    "MeasureSource",
    "Source",
    "SourceRef",
]



from typing import TYPE_CHECKING, Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.hierarchy import Hierarchy
    from pbi_core.ssas.model_tables.relationship import Relationship


class Variation(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/b9dfeb51-cbb6-4eab-91bd-fa2b23f51ca3)
    """

    column: int | None = None  # TODO: pbi says this shouldn't exist
    column_id: int
    default_column_id: int | None = None
    default_hierarchy_id: int
    description: str | None = None
    is_default: bool
    name: str
    relationship_id: int

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.variation)

    def get_column(self) -> "Column":
        """Name is bad to not consistent with other methods because the column field in this entity :(."""
        return self.tabular_model.columns.find(self.column_id)

    def default_column(self) -> "Column | None":
        if self.default_column_id is None:
            return None
        return self.tabular_model.columns.find(self.default_column_id)

    def default_hierarchy(self) -> "Hierarchy":
        return self.tabular_model.hierarchies.find(self.default_hierarchy_id)

    def relationship(self) -> "Relationship":
        return self.tabular_model.relationships.find(self.relationship_id)

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        return LineageNode(
            self,
            lineage_type,
            [
                self.default_hierarchy().get_lineage(lineage_type),
                self.relationship().get_lineage(lineage_type),
            ],
        )



from enum import Enum, IntEnum
from typing import TYPE_CHECKING, Any, Literal
from uuid import UUID

from pydantic import ConfigDict, Field, Json

from pbi_core.lineage.main import LineageNode
from pbi_core.static_files.model_references import ModelColumnReference, ModelMeasureReference

from ._base_node import LayoutNode
from .filters import PageFilter
from .performance import Performance, get_performance
from .selector import Selector
from .sources import Source
from .visual_container import VisualContainer

if TYPE_CHECKING:
    from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel, LocalTabularModel

    from .layout import Layout


class DisplayOption(IntEnum):
    DEPRECATED_DYNAMIC = -1
    """No dynamic page without width or height.
    Deprecated: Use other display options."""
    FIT_TO_PAGE = 0
    """Page is scaled so both width and height fit on the current viewport."""
    FIT_TO_WIDTH = 1
    """Only width is scaled to fit on the current viewport, height will be updated to maintain page aspect ratio."""
    ACTUAL_SIZE = 2
    """No scaling is done - page is centered relative to the report canvas."""
    ACTUAL_SIZE_TOP_LEFT = 3
    """No scaling is done - page is anchored to top-left corder relative to the report canvas.
    Deprecated: Use ActualSize instead."""


class SectionVisibility(IntEnum):
    VISIBLE = 0
    HIDDEN = 1


class SectionConfig(LayoutNode):
    visibility: SectionVisibility = SectionVisibility.VISIBLE
    model_config = ConfigDict(extra="allow")


class BindingType(Enum):
    DEFAULT = "Default"
    """No specific usage of this binding."""
    DRILL_THROUGH = "Drillthrough"
    """Binding to be used as drillthrough."""
    TOOLTIP = "Tooltip"
    """Binding to be used as tooltip page."""


class ReferenceScope(Enum):
    DEFAULT = "Default"
    """Scope is restricted to the report."""
    CROSS_REPORT = "CrossReport"
    """Scope is across reports - for cross-report drillthrough."""


class AcceptsFilterContext(Enum):
    DEFAULT = "Default"
    """Flows filter context"""
    NONE = "None"
    """Additional filter context does not flow to the binding."""


class BindingParameter(LayoutNode):
    name: str
    """Name of the parameter - unique across the report."""
    boundFilter: str | None = None
    """Name of the filter which this parameter affects."""
    asAggregation: bool = False
    """The parameter should be applied when the field of the filter is aggregated."""
    qnaSingleSelectRequired: bool = False
    """Exactly one instance value should be picked as a filter for this parameter."""
    fieldExpr: Source | None = None
    """Field expression for page binding"""


class PageBinding(LayoutNode):
    name: str
    """Name of this binding - unique across the report."""
    type: BindingType
    """Specific usage of this binding (for example drillthrough)."""
    referenceScope: ReferenceScope | None = None
    """What is the scope under which the binding applies."""
    parameters: list[BindingParameter] | None = None
    """Additional parameters to apply when the binding is invoked."""
    acceptsFilterContext: AcceptsFilterContext | None = None
    """Should additional filter context flow when applying the binding."""


class PageInformationProperties(LayoutNode):
    pageInformationName: Any = None
    pageInformationQnaPodEnabled: Any = None
    pageInformationAltName: Any = None
    pageInformationType: Any = None


class PageInformation(LayoutNode):
    selector: Selector | None = None
    """Defines the scope at which to apply the formatting for this object.
    Can also define rules for matching highlighted values and how multiple definitions for the same property should
    be ordered."""
    propeties: PageInformationProperties
    """Describes the properties of the object to apply formatting changes to."""


class PageSizeProperties(LayoutNode):
    pageSizeTypes: Any = None
    pageSizeWidth: Any = None
    pageSizeHeight: Any = None


class PageSize(LayoutNode):
    selector: Selector | None = None
    properties: PageSizeProperties


class PageSizeObjects(LayoutNode):
    pageInformation: list[PageInformation]
    pageSize: list[PageSize]


class BackgroundProperties(LayoutNode):
    color: Any = None
    image: Any = None
    transparency: Any = None


class Background(LayoutNode):
    selector: Selector | None = None
    properties: BackgroundProperties


class DisplayAreaProperties(LayoutNode):
    verticalAlignment: Any = None


class DisplayArea(LayoutNode):
    selector: Selector | None = None
    properties: DisplayAreaProperties


class OutspacePaneProperties(LayoutNode):
    backgroundColor: Any = None
    transparency: Any = None
    foregroundColor: Any = None
    titleSize: Any = None
    searchTextSize: Any = None
    headerSize: Any = None
    fontFamily: Any = None
    border: Any = None
    borderColor: Any = None
    checkboxAndApplyColor: Any = None
    inputBoxColor: Any = None
    width: Any = None


class OutspacePane(LayoutNode):
    selector: Selector | None = None
    properties: OutspacePaneProperties


class FilterCardProperties(LayoutNode):
    backgroundColor: Any = None
    transparency: Any = None
    border: Any = None
    borderColor: Any = None
    foregroundColor: Any = None
    textSize: Any = None
    fontFamily: Any = None
    inputBoxColor: Any = None


class FilterCard(LayoutNode):
    selector: Selector | None = None
    properties: FilterCardProperties


class PageRefreshProperties(LayoutNode):
    show: Any = None
    refreshType: Any = None
    duration: Any = None
    dialogLauncher: Any = None
    measure: Any = None
    checkEvery: Any = None


class PageRefresh(LayoutNode):
    selector: Selector | None = None
    properties: PageRefreshProperties


class PersonalizeVisualProperties(LayoutNode):
    show: Any = None
    perspectiveRef: Any = None
    applyToAllPages: Any = None


class PersonalizeVisual(LayoutNode):
    selector: Selector | None = None
    properties: PersonalizeVisualProperties


class PageFormattingObjects(LayoutNode):
    pageInformation: list[PageInformation]
    pageSize: list[PageSize]
    background: list[Background]
    displayArea: list[DisplayArea]
    outspace: list[Background]  # not a typo, this matches the json schema
    outspacePane: list[OutspacePane]
    filterCard: list[FilterCard]
    pageRefresh: list[PageRefresh]
    personalizeVisuals: list[PersonalizeVisual]


class PageType(Enum):
    DRILL_THROUGH = "Drillthrough"
    """Page to be used as drillthrough."""
    TOOLTIP = "Tooltip"
    """Page to be used as tooltip."""


class PageVisibility(Enum):
    ALWAYS_VISIBLE = "AlwaysVisible"
    """Page is always shown in the pages list"""
    HIDDEN_IN_VIEW_MODE = "HiddenInViewMode"
    """Page is not visible when viewing report in View mode."""


class VisualInteractionFilterType(Enum):
    DEFAULT = "Default"
    """The target visual type determines if it should accept the interaction as a highlight or as a filter."""
    DATA_FILTER = "DataFilter"
    """Data point selection is added as a filter to the target visual."""
    HIGHLIGHT_FILTER = "HighlightFilter"
    """Data point selection is added as a highlight to the target visual."""
    NO_FILTER = "NoFilter"
    """Data point selection is ignored by the target visual."""


class VisualInteraction(LayoutNode):
    source: str
    """Visual name that will be the source of user interaction (selecting data point for example)."""
    target: str
    """Visual name for the target of the interaction (selecting data point for example)."""
    type: VisualInteractionFilterType
    """How should the interaction flow from source to target visual (as highlights, as filter, none)."""


class QuickExploreVisualContainerConfig(LayoutNode):
    name: str
    """Name of the visual - matches the name property in visual.json files"""
    fields: list[Source]
    """Specific data fields used to build this visual from the full set of selected fields"""


class QuickExploreRelatedLayout(LayoutNode):
    version: Literal[1] = 1
    dataTableName: str | None = None
    """If data table is shown, then the name of that visual"""


class QuickExploreCombinationLayout(LayoutNode):
    version: Literal[1] = 1
    dataTableName: str | None = None
    """If data table is shown, then the name of that visual"""


class QuickExploreLayoutContainer(LayoutNode):
    related: QuickExploreRelatedLayout
    """A layout that has 1 hero visual and some related visuals"""
    combination: QuickExploreCombinationLayout
    """Layout that generates visuals purely based on combination of fields
    Deprecated: Use related layout instead."""


class AutoPageGenerationConfig(LayoutNode):
    selectedFields: list[Source]
    """Data fields to use for the auto page generation"""
    visualContainerConfigurations: list[QuickExploreVisualContainerConfig]
    """Visuals already on the page previously generated by the auto-config"""
    layout: QuickExploreLayoutContainer | None = None
    """The specific layout chosen to render the auto-visuals"""


class PageHowCreated(Enum):
    DEFAULT = "Default"
    """Page is generated by user interaction."""
    COPILOT = "Copilot"
    """Page is created by copilot."""


class Annotation(LayoutNode):
    name: str
    """Unique name for the annotation."""
    value: str
    """A value for this annotation."""


class Section(LayoutNode):
    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false
    height: int
    """Height of the page (in pixels) - optional only for 'DeprecatedDynamic' option, required otherwise."""
    width: int
    """Width of the page (in pixels) - optional only for 'DeprecatedDynamic' option, required otherwise."""
    displayOption: DisplayOption
    """Defines how the page is scaled."""
    config: Json[SectionConfig]
    objectId: UUID | None = None
    visualContainers: list[VisualContainer]
    ordinal: int = 0
    filters: Json[list[PageFilter]]
    """Filters that apply to all the visuals on this page - on top of the filters defined for the whole report."""
    displayName: str
    """A user facing name for this page."""
    name: str = Field(max_length=50)
    """A unique identifier for the page across the whole report."""
    id: int | None = None
    pageBinding: PageBinding | None = None
    """Additional metadata defined for how this page is used (tooltip, drillthrough, etc)."""
    objects: PageFormattingObjects | None = None
    """Defines the formatting for different objects on a page."""
    type: PageType | None = None
    """Specific usage of this page (for example drillthrough)."""
    visibility: PageVisibility | None = PageVisibility.ALWAYS_VISIBLE
    """Defines when this page should be visible - by default it is always visible."""
    visualInteractions: list[VisualInteraction] | None = None
    """Defines how data point selection on a specific visual flow (as filters) to other visuals on the page.
    By default it is up-to the visual to apply it either as a cross-highlight or as a filter."""
    autoPageGenerationConfig: AutoPageGenerationConfig | None = None
    """Configuration that was used to automatically generate a page (for example using 'Auto create the report'
    option)."""
    annotations: list[Annotation] | None = None
    """Additional information to be saved (for example comments, readme, etc) for this page."""
    howCreated: PageHowCreated | None = None
    """Source of creation of this page."""

    def pbi_core_name(self) -> str:
        return self.name

    def get_ssas_elements(
        self,
        *,
        include_visuals: bool = True,
        include_filters: bool = True,
    ) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this report page is directly dependent on."""
        ret: set[ModelColumnReference | ModelMeasureReference] = set()
        if include_visuals:
            for viz in self.visualContainers:
                ret.update(viz.get_ssas_elements())
        if include_filters:
            for f in self.filters:
                ret.update(f.get_ssas_elements())
        return ret

    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)

        page_filters = self.get_ssas_elements()
        report_filters = self._parent.get_ssas_elements(include_sections=False)
        entities = page_filters | report_filters
        children_nodes = [ref.to_model(tabular_model) for ref in entities]

        children_lineage = [p.get_lineage(lineage_type) for p in children_nodes if p is not None]
        return LineageNode(self, lineage_type, children_lineage)

    def get_performance(self, model: "LocalTabularModel", *, clear_cache: bool = False) -> list[Performance]:
        """Calculates various metrics on the speed of the visual.

        Current Metrics:
            Total Seconds to Query
            Total Rows Retrieved

        Raises:
            ValueError: If the page does not have any querying visuals.

        """
        commands: list[str] = []
        for viz in self.visualContainers:
            if viz.query is not None:
                command = viz._get_data_command()
                if command is not None:
                    commands.append(command.get_dax(model).dax)
        if not commands:
            msg = "Cannot get performance for a page without any querying visuals"
            raise ValueError(msg)
        return get_performance(model, commands, clear_cache=clear_cache)



import dataclasses
import pathlib
import socket
from typing import Any
from xml.sax.saxutils import escape  # nosec

import jinja2
import psutil

from ._commands import BaseCommands, ModelCommands, RefreshCommands, RenameCommands

COMMAND_DIR: pathlib.Path = pathlib.Path(__file__).parent / "command_templates"

COMMAND_TEMPLATES: dict[str, jinja2.Template] = {
    f.name: jinja2.Template(f.read_text()) for f in COMMAND_DIR.iterdir() if f.is_file()
}
commands: dict[str, dict[str, str]] = {
    folder.name: {f.name: f.read_text() for f in folder.iterdir() if f.is_file()}
    for folder in (COMMAND_DIR / "schema").iterdir()
    if folder.is_dir()
}


class SsasCommands:
    annotation = RenameCommands.new(commands["Annotations"])
    calculation_group = BaseCommands.new(commands["CalculationGroup"])
    calculation_item = RenameCommands.new(commands["CalculationItems"])
    column = RenameCommands.new(commands["Columns"])
    column_permission = BaseCommands.new(commands["ColumnPermissions"])
    culture = RenameCommands.new(commands["Cultures"])
    data_source = RenameCommands.new(commands["DataSources"])
    detail_row_defintion = BaseCommands.new(commands["DetailRowsDefinition"])
    expression = RenameCommands.new(commands["Expressions"])
    extended_property = RenameCommands.new(commands["ExtendedProperties"])
    format_string_definition = BaseCommands.new(commands["FormatStringDefinitions"])
    hierarchy = RenameCommands.new(commands["Hierarchies"])
    kpi = BaseCommands.new(commands["Kpis"])
    level = RenameCommands.new(commands["Levels"])
    linguistic_metadata = BaseCommands.new(commands["LinguisticMetadata"])
    measure = RenameCommands.new(commands["Measures"])
    model = ModelCommands.new(commands["Model"])
    object_translation = BaseCommands.new(commands["ObjectTranslations"])
    partition = RefreshCommands.new(commands["Partitions"])
    perspective_column = BaseCommands.new(commands["PerspectiveColumns"])
    perspective_hierarchy = BaseCommands.new(commands["PerspectiveHierarchies"])
    perspective_measure = BaseCommands.new(commands["PerspectiveMeasures"])
    perspective = RenameCommands.new(commands["Perspectives"])
    perspective_table = BaseCommands.new(commands["PerspectiveTables"])
    query_group = BaseCommands.new(commands["QueryGroups"])
    refresh_policy = BaseCommands.new(commands["RefreshPolicy"])
    relationship = RenameCommands.new(commands["Relationships"])
    role_membership = BaseCommands.new(commands["RoleMemberships"])
    role = RenameCommands.new(commands["Roles"])
    table_permission = BaseCommands.new(commands["TablePermissions"])
    table = RefreshCommands.new(commands["Tables"])
    variation = RenameCommands.new(commands["Variations"])


ROOT_FOLDER = pathlib.Path(__file__).parents[2]
SKU_ERROR = "ImageLoad/ImageSave commands supports loading/saving data for Excel, Power BI Desktop or Zip files. File extension can be only .XLS?, .PBIX or .ZIP."  # noqa: E501


@dataclasses.dataclass
class ServerInfo:
    """Basic information about an SSAS instance from its PID."""

    port: int
    workspace_directory: pathlib.Path


def get_msmdsrv_info(process: psutil.Process) -> ServerInfo | None:
    """Parses ``ServerInfo`` information from PID information.

    Note:
        This function currently assumes that the SSAS Process is called like
        ``pbi_core`` calls it. If you don't include the ``-s`` flag in the command,
        this function will fail

    """

    def check_ports(proc: psutil.Process) -> int | None:
        ports = [
            conn.laddr.port
            for conn in proc.net_connections()
            if conn.status == "LISTEN"
            and conn.family == socket.AF_INET  # to only get the IPV4, not IPV6 version of the connection
        ]
        if len(ports) != 1:
            return None
        return ports[0]

    def check_workspace(proc: psutil.Process) -> pathlib.Path | None:
        try:
            exe_start_command: list[str] = proc.cmdline()
        except psutil.AccessDenied:
            return None

        if "-s" not in exe_start_command:
            return None
        return pathlib.Path(exe_start_command[exe_start_command.index("-s") + 1])

    if process.name() != "msmdsrv.exe":
        return None
    if (port := check_ports(process)) is None:
        return None
    if (workspace_dir := check_workspace(process)) is None:
        return None
    return ServerInfo(port, workspace_dir)


def python_to_xml(text: Any) -> str:
    """Implements basic XML transformation when returning data to SSAS backend.

    Converts:

    - True/False to true/false

    Args:
        text (Any): a value to be sent to SSAS

    Returns:
        str: A stringified, xml-safe version of the value

    """
    if text in {True, False}:
        return str(text).lower()
    if not isinstance(text, str):
        text = str(text)
    return escape(text)


ROW_TEMPLATE = jinja2.Template(
    """
<row xmlns="urn:schemas-microsoft-com:xml-analysis:rowset">
{%- for k, v in fields %}
    <{{k}}>{{v}}</{{k}}>
{%- endfor %}
</row>
""",
)



from typing import TYPE_CHECKING

from .logging import get_logger
from .ssas.model_tables.column import Column
from .ssas.model_tables.measure import Measure
from .ssas.server import BaseTabularModel, LocalTabularModel, get_or_create_local_server
from .static_files import StaticFiles
from .static_files.model_references import ModelColumnReference, ModelMeasureReference

logger = get_logger()

if TYPE_CHECKING:
    from _typeshed import StrPath

    from pbi_core.ssas.model_tables.table import Table


class BaseReport:
    pass


class WorkspaceReport(BaseReport):
    ssas: BaseTabularModel


class LocalReport(BaseReport):
    """An instance of a PowerBI report from a local PBIX file.

    Args:
        static_files (StaticElements): An instance of all the static files (except DataModel) in the PBIX file

    Examples:
        ```python
        from pbi_core import LocalReport

        report = LocalReport.load_pbix("example.pbix")
        report.save_pbix("example_out.pbix")
        ```

    """

    ssas: LocalTabularModel
    """An instance of a local SSAS Server"""

    static_files: StaticFiles
    """Classes representing the static design portions of the PBIX report"""

    def __init__(self, ssas: LocalTabularModel, static_files: StaticFiles) -> None:
        self.ssas = ssas
        self.static_files = static_files

    @staticmethod
    def load_pbix(path: "StrPath", *, kill_ssas_on_exit: bool = True) -> "LocalReport":
        """Creates a ``LocalReport`` instance from a PBIX file.

        Args:
                path (StrPath): The absolute or local path to the PBIX report
                kill_ssas_on_exit (bool, optional): The LocalReport object depends on a ``msmdsrv.exe`` process that is
                    independent of the Python session process. If this function creates a new ``msmdsrv.exe`` instance
                    and kill_ssas_on_exit is true, the process will be killed on exit.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```

        Returns:
                LocalReport: the local PBIX class

        """
        logger.info("Loading PBIX", path=path)
        server = get_or_create_local_server(kill_on_exit=kill_ssas_on_exit)
        ssas = server.load_pbix(path)
        static_files = StaticFiles.load_pbix(path)
        return LocalReport(ssas=ssas, static_files=static_files)

    def save_pbix(self, path: "StrPath", *, sync_ssas_changes: bool = True) -> None:
        """Creates a new PBIX with the information in this class to the given path.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```
               report.save_pbix("example_out.pbix")

        Args:
            path (StrPath): the path (relative or absolute) to save the PBIX to
            sync_ssas_changes (bool, optional): whether to sync changes made in the SSAS model back to the PBIX file

        """
        if sync_ssas_changes:
            self.ssas.sync_to()
        self.ssas.save_pbix(path)
        self.static_files.save_pbix(path)

    def cleanse_ssas_model(self) -> None:
        """Removes all unused tables, columns, and measures in an SSAS model.

        1. Uses the layout to identify all Measures and Columns being used by the report visuals and filters.
        2. Uses SSAS relationships to identify additional columns and tables used for cross-table filtering.
        3. Traces calculation dependencies (on measures and calculated columns) to identify measures and columns used
             to create report fields
        4. Removes any measure/column that is:
            1. Not in results of 1-3
            2. Not part of a system table
        5. Removes any table that has no column/measure used in the report and no active relationship with a reporting
            table
        """
        report_references = self.static_files.layout.get_ssas_elements()
        model_values = (
            [x.to_model(self.ssas) for x in report_references]
            + [relationship.to_column() for relationship in self.ssas.relationships]
            + [relationship.from_column() for relationship in self.ssas.relationships]
        )
        ret: list[Measure | Column] = []
        for val in model_values:
            ret.append(val)
            ret.extend(val.parents(recursive=True))

        used_tables = {x.table() for x in ret}
        used_measures = {
            x
            for x in ret
            if isinstance(
                x,
                Measure,
            )
        }
        used_columns = {x for x in ret if isinstance(x, Column)}

        # In the examples I've seen, there's a table named "DateTableTemplate_<UUID>" that doesn't seem used,
        # but breaks the system when removed
        tables_to_drop = {t for t in self.ssas.tables if t not in used_tables and not t.is_private}
        columns_to_drop = {
            c
            for c in self.ssas.columns
            if c not in used_columns
            and not c.table().name.startswith("DateTableTemplate")
            and c.table() not in tables_to_drop
        }
        affected_tables: dict[Table, list[Column]] = {}
        for c in columns_to_drop:
            t = c.table()
            if t not in tables_to_drop:
                affected_tables.setdefault(t, []).append(c)

        measures_to_drop = {m for m in self.ssas.measures if m not in used_measures}
        # TODO: convert to batch deletion
        for t in tables_to_drop:
            t.delete()
        for c in columns_to_drop:
            c.delete()
        for m in measures_to_drop:
            m.delete()
        for affected_table, table_columns_to_drop in affected_tables.items():
            for partition in affected_table.partitions():
                partition.remove_columns(table_columns_to_drop)


def column_finder(c: Column, reference: ModelColumnReference) -> bool:
    return c.explicit_name == reference.column and c.table().name == reference.table


def measure_finder(m: Measure, reference: ModelMeasureReference) -> bool:
    return m.name == reference.measure and m.table().name == reference.table



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    fillRule: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class LabelsPropertiesHelper(LayoutNode):
    backgroundColor: Expression | None = None
    backgroundTransparency: Expression | None = None
    color: Expression | None = None
    enableBackground: Expression | None = None
    fontSize: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelOrientation: Expression | None = None
    labelPosition: Expression | None = None
    show: Expression | None = None
    showAll: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper
    selector: Selector | None = None


class LegendPropertiesHelper(LayoutNode):
    legendMarkerRendering: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper
    selector: Selector | None = None


class LineStylesPropertiesHelper(LayoutNode):
    lineStyle: Expression | None = None
    markerShape: Expression | None = None
    shadeArea: Expression | None = None
    showMarker: Expression | None = None
    showSeries: Expression | None = None
    stepped: Expression | None = None
    strokeWidth: Expression | None = None


class LineStylesProperties(LayoutNode):
    properties: LineStylesPropertiesHelper
    selector: Selector | None = None


class ValueAxisPropertiesHelper(LayoutNode):
    alignZeros: Expression | None = None
    end: Expression | None = None
    gridlineShow: Expression | None = None
    secEnd: Expression | None = None
    secShow: Expression | None = None
    secStart: Expression | None = None
    start: Expression | None = None
    show: Expression | None = None


class ValueAxisProperties(LayoutNode):
    properties: ValueAxisPropertiesHelper
    selector: Selector | None = None


class SubheaderPropertiesHelper(LayoutNode):
    fontSize: Expression | None = None


class SubheaderProperties(LayoutNode):
    properties: SubheaderPropertiesHelper


class SmallMultiplesLayoutPropertiesHelper(LayoutNode):
    gridLineColor: Expression | None = None
    gridLineStyle: Expression | None = None
    gridLineType: Expression | None = None
    gridPadding: Expression | None = None
    rowCount: Expression | None = None


class SmallMultiplesLayoutProperties(LayoutNode):
    properties: SmallMultiplesLayoutPropertiesHelper


class CategoryAxisPropertiesHelper(LayoutNode):
    axisType: Expression | None = None


class CategoryAxisProperties(LayoutNode):
    properties: CategoryAxisPropertiesHelper


class LineStackedColumnComboChartProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    labels: list[LabelsProperties] | None = None
    legend: list[LegendProperties] | None = None
    lineStyles: list[LineStylesProperties] | None = None
    smallMultiplesLayout: list[SmallMultiplesLayoutProperties] | None = None
    subheader: list[SubheaderProperties] | None = None
    valueAxis: list[ValueAxisProperties] | None = None


class LineStackedColumnComboChart(BaseVisual):
    visualType: str = "lineStackedColumnComboChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: LineStackedColumnComboChartProperties | None = None



import datetime

from git import TYPE_CHECKING
from pydantic import Json, PrivateAttr

from pbi_core.pydantic import BaseValidation
from pbi_core.ssas.model_tables.base import SsasRenameRecord, SsasTable
from pbi_core.ssas.model_tables.enums import ObjectType
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands
from pbi_core.static_files.layout.sources.column import ColumnSource

if TYPE_CHECKING:
    from collections.abc import Callable


class BinSize(BaseValidation):
    value: float
    unit: int


class BinningMetadata(BaseValidation):
    binSize: BinSize


class ExtendedPropertyValue(BaseValidation):
    version: int
    daxTemplateName: str | None = None
    groupedColumns: list[ColumnSource] | None = None
    binningMetadata: BinningMetadata | None = None


class ExtendedProperty(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/5c1521e5-defe-4ba2-9558-b67457e94569)
    """

    object_id: int
    object_type: ObjectType
    name: str
    type: ObjectType
    value: Json[ExtendedPropertyValue]

    modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.extended_property)

    def object(self) -> "SsasTable":
        """Returns the object the property is describing."""
        mapper: dict[ObjectType, Callable[[int], SsasTable]] = {
            ObjectType.MODEL: lambda _x: self.tabular_model.model,
            ObjectType.DATASOURCE: self.tabular_model.data_sources.find,
            ObjectType.TABLE: self.tabular_model.tables.find,
            ObjectType.COLUMN: self.tabular_model.columns.find,
            ObjectType.ATTRIBUTE_HIERARCHY: self.tabular_model.attribute_hierarchies.find,
            ObjectType.PARTITION: self.tabular_model.partitions.find,
            ObjectType.RELATIONSHIP: self.tabular_model.relationships.find,
            ObjectType.MEASURE: self.tabular_model.measures.find,
            ObjectType.HIERARCHY: self.tabular_model.hierarchies.find,
            ObjectType.LEVEL: self.tabular_model.levels.find,
            ObjectType.KPI: self.tabular_model.kpis.find,
            ObjectType.CULTURE: self.tabular_model.cultures.find,
            ObjectType.LINGUISTIC_METADATA: self.tabular_model.linguistic_metadata.find,
            ObjectType.PERSPECTIVE: self.tabular_model.perspectives.find,
            ObjectType.PERSPECTIVE_TABLE: self.tabular_model.perspective_tables.find,
            ObjectType.PERSPECTIVE_HIERARCHY: self.tabular_model.perspective_hierarchies.find,
            ObjectType.PERSPECTIVE_MEASURE: self.tabular_model.perspective_measures.find,
            ObjectType.ROLE: self.tabular_model.roles.find,
            ObjectType.ROLE_MEMBERSHIP: self.tabular_model.role_memberships.find,
            ObjectType.TABLE_PERMISSION: self.tabular_model.table_permissions.find,
            ObjectType.VARIATION: self.tabular_model.variations.find,
            ObjectType.EXPRESSION: self.tabular_model.expressions.find,
            ObjectType.COLUMN_PERMISSION: self.tabular_model.column_permissions.find,
            ObjectType.CALCULATION_GROUP: self.tabular_model.calculation_groups.find,
            ObjectType.QUERY_GROUP: self.tabular_model.query_groups.find,
        }
        return mapper[self.object_type](self.object_id)

    @classmethod
    def _db_command_obj_name(cls) -> str:
        return "ExtendedProperties"



from typing import Annotated, Any, Literal

from pydantic import Discriminator, Tag

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.condition import ConditionType
from pbi_core.static_files.layout.sources import LiteralSource, MeasureSource, Source
from pbi_core.static_files.layout.sources.aggregation import AggregationSource, SelectRef
from pbi_core.static_files.layout.sources.column import ColumnSource


class LiteralExpression(LayoutNode):
    expr: LiteralSource


class MeasureExpression(LayoutNode):
    expr: MeasureSource


class AggregationExpression(LayoutNode):
    expr: AggregationSource


class ThemeDataColor(LayoutNode):
    ColorId: int
    Percent: float


class ThemeExpression(LayoutNode):
    ThemeDataColor: ThemeDataColor


class FillRule(LayoutNode):
    FillRule: "Expression"
    Input: Source


class FillRuleExpression(LayoutNode):
    FillRule: FillRule


class ConditionalCase(LayoutNode):
    Condition: ConditionType
    Value: LiteralSource


class ConditionalSourceHelper(LayoutNode):
    Cases: list[ConditionalCase]


class ConditionalSource(LayoutNode):
    Conditional: ConditionalSourceHelper


class ConditionalExpression(LayoutNode):
    expr: ConditionalSource


def get_subexpr_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        keys = list(v.keys())
        assert len(keys) == 1, f"Expected single key, got {keys}"
        mapper = {
            "ThemeDataColor": "ThemeExpression",
            "Aggregation": "AggregationSource",
            "Literal": "LiteralSource",
            "Measure": "MeasureSource",
            "FillRule": "FillRuleExpression",
            "Conditional": "ConditionalSource",
        }
        if keys[0] in mapper:
            return mapper[keys[0]]
        msg = f"Unknown type: {v.keys()}"
        raise TypeError(msg)
    return v.__class__.__name__


ColorSubExpression = Annotated[
    Annotated[ThemeExpression, Tag("ThemeExpression")]
    | Annotated[LiteralSource, Tag("LiteralSource")]
    | Annotated[MeasureSource, Tag("MeasureSource")]
    | Annotated[FillRuleExpression, Tag("FillRuleExpression")]
    | Annotated[AggregationSource, Tag("AggregationSource")]
    | Annotated[ConditionalSource, Tag("ConditionalSource")],
    Discriminator(get_subexpr_type),
]


class ColorExpression(LayoutNode):
    expr: ColorSubExpression


def get_color_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "expr" in v:
            return "ColorExpression"
        if "Literal" in v:
            return "LiteralSource"
        msg = f"Unknown Color Type: {v.keys()}"
        raise TypeError(msg)
    return v.__class__.__name__


Color = Annotated[
    Annotated[ColorExpression, Tag("ColorExpression")] | Annotated[LiteralSource, Tag("LiteralSource")],
    Discriminator(get_color_type),
]


class SolidExpression(LayoutNode):
    color: Color
    value: LiteralSource | LiteralExpression | None = None  # TODO: explore the cases here more


class SolidColorExpression(LayoutNode):
    solid: SolidExpression


class StrategyExpression(LayoutNode):
    strategy: LiteralExpression | LiteralSource  # TODO: explore the cases here more


class ExtremeColor(LayoutNode):
    color: LiteralSource
    value: LiteralSource


def get_color_helper_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "solid" in v:
            return "SolidExpression"
        if "color" in v:
            return "ExtremeColor"
        msg = f"Unknown class: {v.keys()}"
        breakpoint()
        raise TypeError(msg)
    return v.__class__.__name__


LinearGradient2HelperExtreme = Annotated[
    Annotated[SolidExpression, Tag("SolidExpression")] | Annotated[ExtremeColor, Tag("ExtremeColor")],
    Discriminator(get_color_helper_type),
]


class LinearGradient2Helper(LayoutNode):
    max: SolidExpression
    min: SolidExpression
    nullColoringStrategy: StrategyExpression


class LinearGradient2Expression(LayoutNode):
    linearGradient2: LinearGradient2Helper


class LinearGradient3Helper(LayoutNode):
    max: SolidExpression
    mid: SolidExpression
    min: SolidExpression
    nullColoringStrategy: StrategyExpression


class LinearGradient3Expression(LayoutNode):
    linearGradient3: LinearGradient3Helper


class ResourcePackageItem(LayoutNode):
    PackageName: str
    PackageType: int  # TODO: enum
    ItemName: str


class ResourcePackageAccessExpression(LayoutNode):
    ResourcePackageItem: ResourcePackageItem


class ResourcePackageAccess(LayoutNode):
    expr: ResourcePackageAccessExpression


class ImageKindExpression(LayoutNode):
    kind: Literal["Icon"]
    layout: LiteralExpression
    verticalAlignment: LiteralExpression
    value: ConditionalExpression


# TODO: centralize the expr: Source classes
class SelectRefExpression(LayoutNode):
    expr: SelectRef


class ImageExpressionHelper(LayoutNode):
    name: "Expression"
    scaling: "Expression"
    url: "Expression"


class ImageExpression(LayoutNode):
    image: ImageExpressionHelper


class GeoJsonExpressionHelper(LayoutNode):
    name: "Expression"
    content: "Expression"
    type: "Expression"


class GeoJsonExpression(LayoutNode):
    geoJson: GeoJsonExpressionHelper


class AlgorithmParameter(LiteralSource):
    Name: str


class AlgorithmExpression(LayoutNode):
    algorithm: str
    parameters: list[AlgorithmParameter]


class ExpressionList(LayoutNode):
    exprs: list["Expression"]
    kind: Literal["ExprList"]


class ColumnExpression(LayoutNode):
    expr: ColumnSource


def get_expression(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        mapper = {
            "solid": "SolidColorExpression",
            "linearGradient2": "LinearGradient2Expression",
            "linearGradient3": "LinearGradient3Expression",
            "image": "ImageExpression",
            "geoJson": "GeoJsonExpression",
            "algorithm": "AlgorithmExpression",
        }
        kind_mapper = {
            "Icon": "ImageKindExpression",
            "ExprList": "ExpressionList",
        }
        expr_mapper = {
            "Column": "ColumnExpression",
            "Measure": "MeasureExpression",
            "Literal": "LiteralExpression",
            "Aggregation": "AggregationExpression",
            "ResourcePackageItem": "ResourcePackageAccess",
            "SelectRef": "SelectRefExpression",
        }
        if "kind" in v:
            if v["kind"] in kind_mapper:
                return kind_mapper[v["kind"]]
            msg = f"Unknown kind: {v['kind']}"
            raise ValueError(msg)

        if "expr" in v:
            # Column has multiple keys, so we need to check them
            for k in v["expr"]:
                if k in expr_mapper:
                    return expr_mapper[k]
            msg = f"Unknown expression type: {v['expr']}"
            raise ValueError(msg)

        for key in v:
            if key in mapper:
                return mapper[key]

        msg = f"Unknown class: {v.keys()}"
        raise TypeError(msg)
    return v.__class__.__name__


Expression = Annotated[
    Annotated[LiteralExpression, Tag("LiteralExpression")]
    | Annotated[AlgorithmExpression, Tag("AlgorithmExpression")]
    | Annotated[ColumnExpression, Tag("ColumnExpression")]
    | Annotated[MeasureExpression, Tag("MeasureExpression")]
    | Annotated[AggregationExpression, Tag("AggregationExpression")]
    | Annotated[SolidColorExpression, Tag("SolidColorExpression")]
    | Annotated[LinearGradient2Expression, Tag("LinearGradient2Expression")]
    | Annotated[LinearGradient3Expression, Tag("LinearGradient3Expression")]
    | Annotated[ResourcePackageAccess, Tag("ResourcePackageAccess")]
    | Annotated[ImageKindExpression, Tag("ImageKindExpression")]
    | Annotated[ImageExpression, Tag("ImageExpression")]
    | Annotated[ExpressionList, Tag("ExpressionList")]
    | Annotated[GeoJsonExpression, Tag("GeoJsonExpression")]
    | Annotated[SelectRefExpression, Tag("SelectRefExpression")],
    Discriminator(get_expression),
]



from enum import IntEnum


class RefreshType(IntEnum):
    """From `[`Microsoft doc pages <https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.refreshtype?view=analysisservices-dotnet>_."""

    """
    For all partitions in the specified partition, table, or database, refresh data and recalculate all dependents.
    For a calculation partition, recalculate the partition and all its dependents.
    """
    FULL = 1
    """
    Clear values in this object and all its dependents.
    """
    CLEAR_VALUES = 2
    """
    Recalculate this object and all its dependents, but only if needed.
    This value does not force recalculation, except for volatile formulas.
    """
    CALCULATE = 3
    """
    Refresh data in this object and clear all dependents.
    """
    DATA_ONLY = 4
    """
    If the object needs to be refreshed and recalculated, refresh and recalculate the object and all its dependents.
    Applies if the partition is in a state other than Ready.
    """
    AUTOMATIC = 5
    """
    Append data to this partition and recalculate all dependents.
    This command is valid only for regular partitions and not for calculation partitions.
    """
    ADD = 7
    """
    Defragment the data in the specified table. As data is added to or removed from a table, the dictionaries of each
    column can become polluted with values that no longer exist in the actual column values.
    The defragment option will clean up the values in the dictionaries that are no longer used.
    """
    DEFRAGMENT = 8



from typing import Any, Literal
from uuid import UUID

from ._base import BaseFileModel

base_val = bool | int | str


class Position(BaseFileModel):
    x: int | float
    y: int | float


class Size(BaseFileModel):
    height: int | float
    width: int | float


class Node100(BaseFileModel):
    top: float
    left: float
    width: float
    height: float
    layedOut: bool
    nodeIndex: str


class Node110(BaseFileModel):
    location: Position
    nodeIndex: str
    nodeLineageTag: UUID | None = None
    size: Size
    zIndex: int


class BoundingBoxPosition(BaseFileModel):
    x: float
    y: float


class TableLayout(BaseFileModel):
    boundingBoxHeight: float
    boundingBoxWidth: float
    boundingBoxPosition: BoundingBoxPosition
    nodes: list[Node100]


class DiagramV100(BaseFileModel):
    name: str
    zoomValue: float
    isDefault: bool
    tables: list[str]
    layout: TableLayout


class DiagramV110(BaseFileModel):
    ordinal: int
    scrollPosition: Position
    nodes: list[Node110]
    name: str
    zoomValue: float
    pinKeyFieldsToTop: bool
    showExtraHeaderInfo: bool
    hideKeyFieldsWhenCollapsed: bool
    tablesLocked: bool = False


class DiagramLayoutV100(BaseFileModel):
    version: Literal["1.0.0"]
    diagrams: list[DiagramV100]
    selectedDiagram: str | None = None
    defaultDiagram: str | None = None
    showIntroduceNewModelViewDialog: bool = True


class DiagramLayoutV110(BaseFileModel):
    version: Literal["1.1.0"]
    diagrams: list[DiagramV110]
    selectedDiagram: str | None = None
    defaultDiagram: str | None = None


DiagramLayout = DiagramLayoutV100 | DiagramLayoutV110


def parse_diagram_layout(v: dict[str, Any]) -> DiagramLayout:
    if v["version"] == "1.0.0":
        return DiagramLayoutV100.model_validate(v)
    if v["version"] == "1.1.0":
        return DiagramLayoutV110.model_validate(v)
    msg = f"Unknown Version: {v['version']}"
    raise ValueError(msg)



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.model_tables.table_permission import MetadataPermission, TablePermission
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column


class ColumnPermission(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/10566cbb-390d-470d-b0ff-fc2713277031)
    """

    column_id: int
    metadata_permission: MetadataPermission
    table_permission_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.column_permission)

    def table_permission(self) -> TablePermission:
        return self.tabular_model.table_permissions.find(self.table_permission_id)

    def column(self) -> "Column":
        return self.tabular_model.columns.find(self.column_id)



from pbi_core.ssas.model_tables.base import SsasTable


class RelatedColumnDetail(SsasTable):
    """TBD.

    SSAS spec:
    """

    @classmethod
    def _db_type_name(cls) -> str:
        return "RelatedColumnDetails"



from enum import IntEnum


class NotificationColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class. The following Sub Class Id: Sub Class Name pairs are defined: 0: Proactive Caching Begin 1: Proactive Caching End 2: Flight Recorder Started 3: Flight Recorder Stopped 4: Configuration Properties Updated 5: SQL Trace 6: Object Created 7: Object Deleted 8: Object Altered 9: Proactive Caching Polling Begin 10: Proactive Caching Polling End 11: Flight Recorder Snapshot Begin 12: Flight Recorder Snapshot End 13: Proactive Caching: notifiable object updated 14: Lazy Processing: start processing 15: Lazy Processing: processing complete 16: SessionOpened Event Begin 17: SessionOpened Event End 18: SessionClosing Event Begin 19: SessionClosing Event End 20: CubeOpened Event Begin 21: CubeOpened Event End 22: CubeClosing Event Begin 23: CubeClosing Event End 24: Transaction abort requested
    CURRENTTIME = 2  # Contains the current time of the notification event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Contains the time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Contains the amount of time (in milliseconds) taken by the event.
    INTEGERDATA = 10  # Contains the integer data associated with the notification event. When the EventSubclass column is 8, values are: 1 = Created 2 = Deleted 3 = Changed object's properties 4 = Changed properties of the object's children 6 = Children added 7 = Children deleted 8 = Object fully processed 9 = Object partially processed 10 = Object unprocessed 11 = Object fully optimized 12 = Object partially optimized 13 = Object not optimized
    OBJECTID = 11  # Contains the Object ID for which this notification is issued; this is a string value.
    OBJECTTYPE = 12  # Contains the object type associated with the notification event.
    OBJECTNAME = 13  # Contains the object name associated with the notification event.
    OBJECTPATH = 14  # Contains the object path associated with the notification event. The path is returned as a comma-separated list of parents, starting with the object's parent.
    OBJECTREFERENCE = 15  # Contains the object reference for the progress report end event. The object reference is encoded as XML by all parents by using tags to describe the object.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the notification event.
    DATABASENAME = 28  # Contains the name of the database in which the notification event occurred.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    SESSIONID = 39  # Contains the session ID associated with the notification event.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the notification event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the notification event.
    SERVERNAME = 43  # Contains the name of the Analysis Services instance on which the notification event occurred.
    REQUESTPROPERTIES = 45  # Contains the properties of the XMLA request.
    ACTIVITYID = 46
    REQUESTID = 47


class UserDefinedColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # A specific user event subclass that provides additional information about each event class.
    CURRENTTIME = 2  # Contains the current time of the notification event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    INTEGERDATA = 10  # A specific user defined event information.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the notification event.
    DATABASENAME = 28  # Contains the name of the database in which the notification event occurred.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    SESSIONID = 39  # Contains the session ID associated with the notification event.
    NTCANONICALUSERNAME = 40  # Contains the Windows user name associated with the notification event. The user name is in canonical form. For example, engineering.microsoft.com/software/user.
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the notification event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the notification event.
    SERVERNAME = 43  # Contains the name of the Analysis Services instance on which the notification event occurred.
    ACTIVITYID = 46
    REQUESTID = 47



from enum import IntEnum


class ExistingConnectionColumns(IntEnum):
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Unique connection ID.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SPID = 41  # Server process ID. This uniquely identifies a user session. This directly corresponds to the session GUID used by XML/A.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class ExistingSessionColumns(IntEnum):
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    CPUTIME = 6  # Amount of CPU time (in milliseconds) used by the event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Server process ID. This uniquely identifies a user session. This directly corresponds to the session GUID used by XML/A.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    REQUESTPROPERTIES = 45  # XMLA request properties.
    ACTIVITYID = 46
    REQUESTID = 47


class SessionInitializeColumns(IntEnum):
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Server process ID. This uniquely identifies a user session. This directly corresponds to the session GUID used by XML/A.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    REQUESTPROPERTIES = 45  # XMLA request properties.
    ACTIVITYID = 46
    REQUESTID = 47





from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class DataBarsProperties(LayoutNode):
    axisColor: Expression
    hideText: Expression
    negativeColor: Expression
    positiveColor: Expression
    reverseDirection: Expression


class ColumnFormattingProperties(LayoutNode):
    alignment: Expression | None = None
    backColor: Expression | None = None
    dataBars: DataBarsProperties | None = None
    fontColor: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelPrecision: Expression | None = None
    styleHeader: Expression | None = None
    styleValues: Expression | None = None


class ColumnFormatting(LayoutNode):
    properties: ColumnFormattingProperties
    selector: Selector | None = None


class TotalProperties(LayoutNode):
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    outline: Expression | None = None
    totals: Expression | None = None


class Total(LayoutNode):
    properties: TotalProperties
    selector: Selector | None = None


class ValuesProperties(LayoutNode):
    backColor: Expression | None = None
    backColorPrimary: Expression | None = None
    backColorSecondary: Expression | None = None
    fontColor: Expression | None = None
    fontColorPrimary: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    outline: Expression | None = None
    urlIcon: Expression | None = None
    wordWrap: Expression | None = None


class Values(LayoutNode):
    properties: ValuesProperties
    selector: Selector | None = None


class ColumnHeadersProperties(LayoutNode):
    alignment: Expression | None = None
    autoSizeColumnWidth: Expression | None = None
    backColor: Expression | None = None
    bold: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    outline: Expression | None = None
    outlineStyle: Expression | None = None
    wordWrap: Expression | None = None


class ColumnHeaders(LayoutNode):
    properties: ColumnHeadersProperties
    selector: Selector | None = None


class ColumnWidthProperties(LayoutNode):
    value: Expression


class ColumnWidth(LayoutNode):
    properties: ColumnWidthProperties
    selector: Selector | None = None


class GridProperties(LayoutNode):
    gridHorizontal: Expression | None = None
    gridHorizontalColor: Expression | None = None
    gridHorizontalWeight: Expression | None = None
    gridVertical: Expression | None = None
    gridVerticalColor: Expression | None = None
    gridVerticalWeight: Expression | None = None
    imageHeight: Expression | None = None
    outlineColor: Expression | None = None
    outlineWeight: Expression | None = None
    rowPadding: Expression | None = None
    textSize: Expression | None = None


class Grid(LayoutNode):
    properties: GridProperties
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    pass


class General(LayoutNode):
    properties: GeneralProperties


class TableChartColumnProperties(LayoutNode):
    columnFormatting: list[ColumnFormatting] | None = None
    columnHeaders: list[ColumnHeaders] | None = None
    columnWidth: list[ColumnWidth] | None = None
    general: list[General] | None = None
    grid: list[Grid] | None = None
    total: list[Total] | None = None
    values: list[Values] | None = None


class TableChart(BaseVisual):
    visualType: str = "tableEx"
    objects: TableChartColumnProperties | None = None
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None



from enum import IntEnum
from typing import TYPE_CHECKING

from ._base_node import LayoutNode

if TYPE_CHECKING:
    from .layout import Layout


class ResourcePackageItemType(IntEnum):
    JS = 0
    CSS = 1
    PNG = 3
    PBIVIZ = 5
    NA = 100
    TOPO = 200
    JSON2 = 201
    JSON = 202


class ResourcePackageItem(LayoutNode):
    _parent: "ResourcePackageDetails"  # pyright: ignore reportIncompatibleVariableOverride=false
    name: str | None = None
    path: str
    type: ResourcePackageItemType
    resourcePackageId: int | None = None
    resourcePackageItemBlobInfoId: int | None = None
    id: int | None = None


class ResourcePackageDetailsType(IntEnum):
    JS = 0
    CUSTOM_THEME = 1
    BASE_THEME = 2


class ResourcePackageDetails(LayoutNode):
    _parent: "ResourcePackage"  # pyright: ignore reportIncompatibleVariableOverride=false
    disabled: bool = False
    items: list[ResourcePackageItem] = []
    type: ResourcePackageDetailsType
    name: str
    id: int | None = None
    reportId: int | None = None


class ResourcePackage(LayoutNode):
    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false
    resourcePackage: ResourcePackageDetails



from enum import Enum
from typing import TYPE_CHECKING, Literal

from pydantic import Json

from pbi_core.lineage.main import LineageNode
from pbi_core.static_files.model_references import ModelColumnReference, ModelMeasureReference

from ._base_node import LayoutNode
from .bookmark import LayoutBookmarkChild
from .filters import GlobalFilter
from .pod import Pod
from .resource_package import ResourcePackage
from .section import Section
from .visuals.base import FilterSortOrder
from .visuals.properties.base import Expression

if TYPE_CHECKING:
    from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel


class LayoutOptimization(Enum):
    DESKTOP = 0
    MOBILE = 1


class PublicCustomVisual(LayoutNode):
    pass  # TODO: find an example where this occurs


class ThemeVersionInfo(LayoutNode):
    visual: str
    page: str
    report: str


class ThemeResourcePackageType(Enum):
    REGISTERED_RESORCES = 1
    SHARED_RESOURCES = 2


class ThemeInfo(LayoutNode):
    name: str
    version: ThemeVersionInfo | str
    type: ThemeResourcePackageType  # def an enum


class ThemeCollection(LayoutNode):
    baseTheme: ThemeInfo | None = None
    customTheme: ThemeInfo | None = None


class Settings(LayoutNode):
    allowChangeFilterTypes: bool = True
    allowDataPointLassoSelect: bool = False
    exportDataMode: int = 0  # def an enum
    hideVisualContainerHeader: bool = False
    isPersistentUserStateDisabled: bool = False
    useEnhancedTooltips: bool = True
    useNewFilterPaneExperience: bool = True
    useStylableVisualContainerHeader: bool = True
    useCrossReportDrillthrough: bool = False
    defaultFilterActionIsDataFilter: bool = False
    disableFilterPaneSearch: bool = False
    allowInlineExploration: bool = True
    optOutNewFilterPaneExperience: bool = False
    enableDeveloperMode: bool = False
    filterPaneHiddenInEditMode: bool = False
    queryLimitOption: int = 6
    useDefaultAggregateDisplayName: bool = True


class SlowDataSourceSettings(LayoutNode):
    isCrossHighlightingDisabled: bool
    isFieldWellButtonEnabled: bool
    isFilterSelectionsButtonEnabled: bool
    isSlicerSelectionsButtonEnabled: bool
    isApplyAllButtonEnabled: bool = False


class OutspacePaneProperties(LayoutNode):
    visible: Expression | None = None
    expanded: Expression | None = None


class OutspacePane(LayoutNode):
    properties: OutspacePaneProperties


class ConfigSectionProperties(LayoutNode):
    verticalAlignment: Expression | None = None


class ConfigSection(LayoutNode):
    properties: ConfigSectionProperties


class LayoutProperties(LayoutNode):
    outspacePane: list[OutspacePane] | None = None
    section: list[ConfigSection] | None = None


class LayoutConfig(LayoutNode):
    linguisticSchemaSyncVersion: int | None = None
    defaultDrillFilterOtherVisuals: bool = True
    bookmarks: list[LayoutBookmarkChild] | None = None
    activeSectionIndex: int
    themeCollection: ThemeCollection
    slowDataSourceSettings: SlowDataSourceSettings | None = None
    settings: Settings | None = None
    version: float  # looks like a float
    objects: LayoutProperties | None = None
    filterSortOrder: FilterSortOrder = FilterSortOrder.NA


class Layout(LayoutNode):
    """Represents the layout of a Power BI report, including sections, filters, and other properties."""

    id: int = -3
    reportId: int = -1
    filters: Json[list[GlobalFilter]] = []
    resourcePackages: list[ResourcePackage]
    sections: list[Section]
    config: Json[LayoutConfig]
    layoutOptimization: LayoutOptimization
    theme: str | None = None
    pods: list[Pod] = []
    publicCustomVisuals: list[PublicCustomVisual] = []
    _xpath = []

    def pbi_core_name(self) -> str:  # noqa: PLR6301
        return "Layout"

    def get_ssas_elements(
        self,
        *,
        include_sections: bool = True,
        include_filters: bool = True,
    ) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this report is directly dependent on."""
        ret: set[ModelColumnReference | ModelMeasureReference] = set()
        if include_filters:
            for f in self.filters:
                ret.update(f.get_ssas_elements())
        if include_sections:
            for s in self.sections:
                ret.update(s.get_ssas_elements())
        return ret

    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)

        entities = self.get_ssas_elements()
        children_nodes = [ref.to_model(tabular_model) for ref in entities]

        children_lineage = [p.get_lineage(lineage_type) for p in children_nodes if p is not None]
        return LineageNode(self, lineage_type, children_lineage)




from ._base_node import LayoutNode
from .sources import Source


class Information(LayoutNode):
    method: str
    disabled: bool = False


class ExpansionStateLevel(LayoutNode):
    queryRefs: list[str]
    isCollapsed: bool = False
    isPinned: bool = False
    isLocked: bool = False
    identityKeys: list[Source] | None = None
    identityValues: list[None] | None = None
    AIInformation: Information | None = None


class ExpansionStateChild(LayoutNode):
    isToggled: bool = False
    identityValues: list[Source]
    children: list["ExpansionStateChild"] | None = None


class ExpansionStateRoot(LayoutNode):
    identityValues: list[None] | None = None
    isToggled: bool = False
    children: list[ExpansionStateChild] | None = None


class ExpansionState(LayoutNode):
    roles: list[str]
    isToggled: bool = False
    levels: list[ExpansionStateLevel] | None = None
    root: ExpansionStateRoot | None = None



from typing import Any, ClassVar, Literal

import pydantic
from bs4 import BeautifulSoup
from structlog import get_logger

from pbi_core.lineage import LineageNode
from pbi_core.pydantic.main import BaseValidation
from pbi_core.ssas.model_tables._group import IdBase
from pbi_core.ssas.server._commands import Command
from pbi_core.ssas.server.tabular_model import BaseTabularModel
from pbi_core.ssas.server.utils import ROW_TEMPLATE, python_to_xml

logger = get_logger()


class SsasTable(BaseValidation, IdBase):
    tabular_model: BaseTabularModel
    _read_only_fields: ClassVar[tuple[str, ...]] = ()

    _db_field_names: ClassVar[dict[str, str]] = {}
    _repr_name_field: str = "name"

    @classmethod
    def _db_type_name(cls) -> str:
        return cls.__name__

    @classmethod
    def _db_command_obj_name(cls) -> str:
        """Returns the name of the object expected by their XMLA commands.

        Generally a simple pluralization, but occasionally different in subclasses.
        """
        return cls.__name__ + "s"

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report.

        Uses the _repr_name_field to determine the field to use.
        Defaults to self.name
        """
        return str(getattr(self, self._repr_name_field))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id}: {self.pbi_core_name()})"

    @pydantic.model_validator(mode="before")
    def to_snake_case(cls: "SsasTable", raw_values: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        """Converts SnakeCase to snake_case.

        If a "special_cases" example appears, that transformation is applied.
        If the first character is capitalized, it is lower cased
        If any other character is capitalized, it is lower cased and prefixed with a "_"
        """

        def update_char(char: str, prev_char: str) -> str:
            if char.isupper() and prev_char.islower() and prev_char != "_":
                return f"_{char.lower()}"
            return char.lower()

        def case_helper(field_name: str) -> str:
            SPECIAL_CASES = {  # noqa: N806
                "owerBI": "owerbi",
                "KPIID": "KpiId",
            }
            for old_segment, new_segment in SPECIAL_CASES.items():
                field_name = field_name.replace(old_segment, new_segment)
            return "".join(
                update_char(curr, prev) for prev, curr in zip(" " + field_name[:-1], field_name, strict=False)
            ).strip("_")

        ret: dict[str, Any] = {}
        for field_name, field_value in raw_values.items():
            formatted_field_name = case_helper(field_name)
            if formatted_field_name != field_name:
                cls._db_field_names[formatted_field_name] = field_name
            ret[formatted_field_name] = field_value
        return ret

    def query_dax(self, query: str, db_name: str | None = None) -> None:
        """Helper function to remove the ``.tabular_model.server`` required to run a DAX query from an SSAS element."""
        self.tabular_model.server.query_dax(query, db_name=db_name)

    def query_xml(self, query: str, db_name: str | None = None) -> BeautifulSoup:
        """Helper function to remove the ``.tabular_model.server`` required to run an XML query from an SSAS element."""
        return self.tabular_model.server.query_xml(query, db_name)

    @staticmethod
    def _get_row_xml(values: dict[str, Any], command: Command) -> str:
        fields: list[tuple[str, str]] = []
        for field_name, field_value in values.items():
            if field_name not in command.field_order:
                continue
            if field_value is None:
                continue
            fields.append((field_name, python_to_xml(field_value)))
        fields = command.sort(fields)
        return ROW_TEMPLATE.render(fields=fields)

    @staticmethod
    def render_xml_command(values: dict[str, Any], command: Command, db_name: str) -> str:
        """XMLA commands: create/alter/delete/rename/refresh.

        Commands are generally in the form:
        <batch>
            <create/alter...>
                <db>
            </create/alter...>
            <entity-schema.../>
            <records.../>
        </batch>

        Entity schemas can be found at `pbi_core/ssas/server/command_templates/schema`
        """
        xml_row = SsasTable._get_row_xml(values, command)
        xml_entity_definition = command.entity_template.render(rows=xml_row)
        return command.base_template.render(db_name=db_name, entity_def=xml_entity_definition)

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        """Creates a lineage node tracking the data parents/children of a record."""
        return LineageNode(self, lineage_type)

    def __hash__(self) -> int:
        return hash(self.id)

    def xml_fields(self) -> dict[str, Any]:
        return {
            self._db_field_names.get(k, k): v for k, v in self.model_dump().items() if k not in self._read_only_fields
        }

    def modification_hash(self) -> int:
        """Returns a hash representing the current state of the object.

        Used to identify when the object has changed from it's state in the SSAS instance.
        By default, uses the id field, which means that no fields are tracked for changes.
        """
        return self.id
